from pathlib import Path
from typing import Tuple, Union, Dict, List

import numpy as np
from ase import Atoms
from pymatgen.analysis.local_env import NearNeighbors
from pymatgen.core import Structure, Molecule
from pymatgen.io.ase import AseAtomsAdaptor
from pymatgen.optimization.neighbors import find_points_in_spheres

from featurebox.featurizers.base_transform import BaseFeature
from featurebox.featurizers.envir.desc_env import DesDict
from featurebox.featurizers.envir.local_env import NNDict
from featurebox.utils.look_json import get_marked_class
from featurebox.utils.predefined_typing import StructureOrMolecule, StructureOrMoleculeOrAtoms

MODULE_DIR = Path(__file__).parent.parent.parent.absolute()


def universe_refine_des(d: Dict, fill_size=10, **kwargs):
    """
    Change each center atoms has fill_size neighbors.
    More neighbors would be abandoned.
    Insufficient neighbors would be duplicated.

    Args:
        d: dict, dict of descriptor. at lest contain "x" and "dxdr"
        fill_size: int

    Returns:
        (center_indices,center_prop, neighbor_indices, images, distances)\n
        center_indices: np.ndarray 1d(N,1).\n
        center_prop: np.ndarray 1d(N,l_c).\n
        neighbor_indices: np.ndarray 2d(N,fill_size).\n
        images: np.ndarray 2d(N,fill_size,l).\n
        distance: np.ndarray 2d(N,fill_size).

    """
    center, dxdr, seq = d["x"], d["dxdr"], d.get("seq")
    atom_len = center.shape[0]
    dxdr_len = dxdr.shape[0]

    if dxdr.ndim == 3:
        seq0 = seq[:, 0]

        uni = [len(seq[np.where(seq0 == i)]) for i in range(np.max(seq0) + 1)]
        if len(set(uni)) == 1:
            fill_size1 = int(dxdr_len / atom_len)
            dxdr_new = np.reshape(dxdr, (atom_len, fill_size1, -1))
            seq_len = seq.shape[0]
            fill_size2 = int(seq_len / atom_len)
            # assert fill_size2==fill_size1
            seq_new = np.reshape(seq[:, 1], (atom_len, fill_size2))
            left = fill_size - uni[0]
            if left == 0:
                pass
            elif left > 0:
                seq_new = np.concatenate((seq_new, np.array([seq_new[:, -1]] * left).T), axis=1)
                dxdr_new = np.concatenate((dxdr_new, np.array([dxdr_new[:, -1]] * left).transpose(1, 0, 2)), axis=1)

            else:
                seq_new = seq_new[:, :fill_size]
                dxdr_new = dxdr_new[:, :fill_size]

        else:
            seq_split = [list(seq[:, 1][np.where(seq0 == i)]) for i in range(np.max(seq0) + 1)]

            left = [fill_size - i for i in uni]
            seq_new = []
            dxdr_new = []
            for lefti, seqsi in zip(left, seq_split):
                if lefti == 0:
                    dxdri = dxdr[seqsi]
                elif lefti > 0:
                    seqsi.extend([seqsi[-1]] * lefti)
                    dxdri = dxdr[seqsi]
                else:
                    seqsi = seqsi[:fill_size]  # ? it is not loosely
                    dxdri = dxdr[:fill_size]

                seq_new.append(seqsi)
                dxdr_new.append(dxdri)
            seq_new = np.array(seq_new)
            dxdr_new = np.array(dxdr_new)

    elif dxdr.ndim == 4:
        dxdr_new = np.reshape(dxdr, (dxdr.shape[0], dxdr.shape[1], -1))
        seq_new = np.repeat(np.arange(atom_len).reshape(-1, 1), dxdr.shape[1], axis=1).T
        left = fill_size - atom_len
        if left == 0:
            pass
        elif left > 0:
            seq_new = np.concatenate((seq_new, np.array([seq_new[:, -1]] * left).T), axis=1)
            dxdr_new = np.concatenate((dxdr_new, np.array([dxdr_new[:, -1]] * left).transpose(1, 0, 2)), axis=1)

        else:
            seq_new = seq_new[:, :fill_size]
            dxdr_new = dxdr_new[:, :fill_size]
    else:
        raise NotImplementedError("Unrecognized data format")

    assert seq_new.shape[0] == center.shape[0]
    assert seq_new.shape[1] == dxdr_new.shape[1]
    return np.array(range(atom_len)), center, seq_new, dxdr_new, None


def universe_refine_nn(center_indices, neighbor_indices, distances, vectors=None, center_prop=None, ele_numbers=None,
                       fill_size=5,
                       dis_sort=False,
                       **kwargs):
    """
    Change each center atoms has fill_size neighbors.
    More neighbors would be abandoned.
    Insufficient neighbors would be duplicated.

    Args:
        center_indices: np.ndarray 1d
        neighbor_indices: np.ndarray 1d
        distances: np.ndarray 1d
        vectors: np.ndarray 2d
        fill_size: float
        dis_sort:bool
            sort neighbors with distance.

    Returns:
        (center_indices,center_indices,  neighbor_indices, images, distances)\n
        center_indices: np.ndarray 1d(N,1).\n
        center_prop: np.ndarray 1d(N,l_c).\n
        neighbor_indices: np.ndarray 2d(N,fill_size).\n
        images: np.ndarray 2d(N,fill_size,l).\n
        distance: np.ndarray 2d(N,fill_size,1).

    where l, and l_c >= 1
    """
    _ = ele_numbers

    neis = []
    diss = []
    vecs = []
    cen = np.array(sorted(set(center_indices.tolist())))
    for i in cen:
        # try:
        cidx = np.where(center_indices == i)[0]
        nei = neighbor_indices[cidx]
        disi = distances[cidx]
        if vectors is not None:
            vec = vectors[cidx, :]
        else:
            vec = None

        if dis_sort:
            neidx = np.argsort(disi)
            nei = nei[neidx]
            disi = disi[neidx]
            if vec is not None:
                vec = vec[neidx, :]

        if len(disi) >= fill_size:
            nei = nei[:fill_size]
            disi = disi[:fill_size]
            if vec is not None:
                vec = vec[:fill_size, :]
        else:
            while len(disi) < fill_size:
                nei = np.append(nei, nei[-1])
                disi = np.append(disi, disi[-1])
                if vec is not None:
                    vec = np.concatenate((vec, vec[-1, :].reshape(1, -1)), axis=0)

        neis.append(nei)
        diss.append(disi)

        if vectors is not None:
            vecs.append(vec)
    if vectors is not None:
        vecs = np.array(vecs)
    else:
        vecs = []
    if center_prop is None:
        return np.array(cen).ravel(), np.array(cen).reshape(-1, 1), np.array(neis), vecs, np.array(diss)[
            ..., np.newaxis]
    else:
        return np.array(cen).ravel(), center_prop, np.array(neis), vecs, np.array(diss)[..., np.newaxis]


# def perovskite_refine_nn(center_indices, neighbor_indices, distances, vectors=None, center_prop=None,
#                          ele_numbers=None, fill_size=5, dis_sort=False, **kwargs):
#     """
#     Change each center atoms has fill_size neighbors.
#     More neighbors would be abandoned.
#     Insufficient neighbors would be duplicated.
#
#     Args:
#         center_indices: np.ndarray 1d
#         neighbor_indices: np.ndarray 1d
#         distances: np.ndarray 1d
#         vectors: np.ndarray 2d
#         fill_size: float
#         dis_sort:bool
#             sort neighbors with distance.
#
#     Returns:
#         (center_indices,center_indices,  neighbor_indices, images, distances)\n
#         center_indices: np.ndarray 1d(N,1).\n
#         center_prop: np.ndarray 1d(N,l_c).\n
#         neighbor_indices: np.ndarray 2d(N,fill_size).\n
#         images: np.ndarray 2d(N,fill_size,l).\n
#         distance: np.ndarray 2d(N,fill_size,1).
#
#     where l, and l_c >= 1
#     """
#     # assert center_prop == None
#     #
#     # # todo
#     #
#     # return np.array(cen).ravel(), np.array(cen).reshape(-1, 1), np.array(neis), vecs, np.array(diss)[
#     #         ..., np.newaxis]
#     return


class _BaseEnvGet(BaseFeature):
    """
    Base class return features of each atoms, meanwhile it re-back bond feature.
    """

    def __init__(self, n_jobs: int, on_errors: str = 'raise', return_type: str = 'any',
                 batch_calculate: bool = False,
                 batch_size: int = 30):
        super().__init__(n_jobs=n_jobs, on_errors=on_errors, return_type=return_type,
                         batch_calculate=batch_calculate,
                         batch_size=batch_size)
        self.adaptor_dict = {}  # Dynamic assignment
        self.adaptor = AseAtomsAdaptor()

    def ase_to_pymatgen(self, atom: Atoms, prop_dict=None):
        """ase_to_pymatgen"""
        if not prop_dict:
            prop_dict = self.adaptor_dict if self.adaptor_dict else {}

        structure = self.adaptor.get_structure(atom)
        if prop_dict != {}:
            [structure.__setattr__(k, v) for k, v in prop_dict.items()]
        return structure

    def pymatgen_to_ase(self, structure: Structure, prop_dict=None):
        """pymatgen_to_ase"""
        if not prop_dict:
            prop_dict = self.adaptor_dict if self.adaptor_dict else {}

        atoms = self.adaptor.get_atoms(structure)
        if prop_dict != {}:
            [atoms.__setattr__(k, v) for k, v in prop_dict.items()]
        return atoms


class BaseDesGet(_BaseEnvGet):
    """
    Get atoms features from Pymatgen.structure.
    Though, the class return features of each atoms, meanwhile it re-back bond feature.

    And get following :

    ``center_indices``:np.ndarray of shape(n,)
        center indexes.
    ``center_indices``:np.ndarray of shape(n,l_c)
        center properties.
    ``neighbor_indices``:np.ndarray of shape(n,fill_size)
        neighbor_indexes for each center_index.
        `fill_size` is the parameter of `refine` function.
    ``images``:np.ndarray of shape(n,lb>=3)
        offset vector in 3 orientations or more bond properties.
    ``distances``:np.ndarray of shape(n,fill_size)
        distance of neighbor_indexes for each center_index.
    """

    def __init__(self, nn_strategy="SOAP", refine: str = None,
                 refined_strategy_param: Dict = None,
                 numerical_tol=1e-8, pbc: List[int] = None, cutoff=None, cut_off_name=None):
        """

        Parameters
        ----------
        nn_strategy:
            pyXtelff descriptors, which has calculate method.
        refine:str
            sort method for neighbors of each atom.
            all the refine_method should return 5 result.
            See Also:
                :func:`refine_method_nn`
        refined_strategy_param:dict
            parameters for refine
        numerical_tol:float
            numerical_tol
        pbc:list
            periodicity in 3 direction
            3-length list,each one is 1 or 0. such as [0,0,0],The 1 mean in this direction is with periodicity.
        cutoff:float

        """

        super().__init__(n_jobs=1, on_errors='raise', return_type='any')
        if nn_strategy is None:
            nn_strategy = "SOAP"
        self.nn_strategy = get_marked_class(nn_strategy, DesDict)
        assert hasattr(self.nn_strategy, "calculate"), "The strategy must have calculate method"
        if refine is None:
            self.refine = universe_refine_des
        else:
            self.refine = after_treatment_func_map_des[refine]
        self.refined_strategy_param = refined_strategy_param if isinstance(refined_strategy_param, dict) else {}

        if cut_off_name:
            if cutoff:
                setattr(self.nn_strategy, cut_off_name, cutoff)
            self.cut_off_name = cut_off_name

        else:
            if cutoff is not None:
                if hasattr(nn_strategy, "cutoff"):
                    self.nn_strategy.cutoff = cutoff
                    self.cut_off_name = "cutoff"

                if hasattr(nn_strategy, "Rc"):
                    self.nn_strategy.Rc = cutoff
                    self.cut_off_name = "Rc"

                if hasattr(nn_strategy, "rcut"):
                    self.nn_strategy.rcut = cutoff
                    self.cut_off_name = "rcut"
                else:
                    raise TypeError("Can't find cut radius name")
            else:
                self.cut_off_name = cut_off_name

        self.cutoff = cutoff
        self.numerical_tol = numerical_tol
        self.pbc = pbc

    def convert(self, structure):
        """

        Args:
            structure:Structure,pymatgen Structure

        Returns:
            center_indices,center_prop,neighbor_indices,images,distances
        """

        return self.get_graphs_with_strategy_merge_sort(structure)

    def _calculate(self, d):
        if isinstance(d, Structure):
            d = self.pymatgen_to_ase(d)

        d = self.nn_strategy.calculate(d)
        return d

    def get_graphs_with_strategy_merge_sort(self, structure: StructureOrMoleculeOrAtoms
                                            ) -> Tuple[np.ndarray]:
        """For get bond distance with different strategy, for different nn_stagy could be rewrite"""

        try:
            result_dict = self._calculate(structure)
            result = self.refine(result_dict)
        except ValueError as e:
            print(e)
            print("For {}, There is no neighbor in cutoff {} A, try with cutoff {} A for this structure.".format(
                str(structure.composition), getattr(self.nn_strategy, self.cut_off_name),
                getattr(self.nn_strategy, self.cut_off_name) + 2)
            )
            setattr(self.nn_strategy, self.cut_off_name, getattr(self.nn_strategy, self.cut_off_name) + 2)
            result_dict = self._calculate(structure)
            result = self.refine(result_dict)

        return result


class BaseNNGet(BaseFeature):
    """
    Get properties from Pymatgen.structure.
    And get following :

    ``center_indices``:np.ndarray of shape(n,)
        center indexes.
    ``center_indices``:np.ndarray of shape(n,l_c)
        center properties.
    ``neighbor_indices``:np.ndarray of shape(n,fill_size)
        neighbor_indexes for each center_index.
        `fill_size` is the parameter of `refine` function.
    ``images``:np.ndarray of shape(n,lb>=3)
        offset vector in 3 orientations or more bond properties.
    ``distances``:np.ndarray of shape(n,fill_size)
        distance of neighbor_indexes for each center_index.

    """

    def __init__(self, nn_strategy: Union[NearNeighbors] = "UserVoronoiNN", refine: str = None,
                 refined_strategy_param: Dict = None,
                 numerical_tol=1e-8, pbc: List[int] = None, cutoff=5.0):
        """

        Parameters
        ----------
        nn_strategy: Union[NearNeighbors]
            search method for local_env for each atom.
        refine:str
            sort method for neighbors of each atom.
            See Also:
                :func:`refine_method_nn`
        refined_strategy_param:dict
            parameters for refine
        numerical_tol:float
            numerical_tol
        pbc:list
            only for find "find_points_in_spheres".
            periodicity in 3 direction
            3-length list,each one is 1 or 0. such as [0,0,0],The 1 mean in this direction is with periodicity.
        cutoff:
            if offered, the nn_strategy would be neglect and find neighbors using
                ``find_points_in_spheres`` in pymatgen.
        """

        super().__init__(n_jobs=1, on_errors='raise', return_type='any')
        self.nn_strategy = get_marked_class(nn_strategy, NNDict)
        if refine is None:
            self.refine = universe_refine_nn
        else:
            self.refine = after_treatment_func_map_nn[refine]
        self.refined_strategy_param = refined_strategy_param if isinstance(refined_strategy_param, dict) else {}
        self.cutoff = cutoff
        self.numerical_tol = numerical_tol
        self.pbc = pbc

    def convert(self, structure: StructureOrMolecule) -> Tuple[
        np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """

        Args:
            structure:Structure,pymatgen Structure

        Returns:
            center_indices,center_prop,neighbor_indices,images,distances
        """
        if self.nn_strategy == "find_points_in_spheres":
            return self.get_graphs_within_cutoff_merge_sort(structure)
        else:
            return self.get_graphs_with_strategy_merge_sort(structure)

    def _within_cutoff(self,
                       structure: StructureOrMolecule, cutoff: float = 5.0, numerical_tol: float = 1e-8
                       ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """
        Get graph representations from structure within cutoff.

        Args:
            structure (pymatgen Structure or molecule)
            cutoff (float): cutoff radius
            numerical_tol (float): numerical tolerance

        Returns:
            center_indices, neighbor_indices, images, distances
        """
        if isinstance(structure, Structure):
            lattice_matrix = np.ascontiguousarray(np.array(structure.lattice.matrix), dtype=float)
            pbc = np.array([1, 1, 1], dtype=int)
        elif isinstance(structure, Molecule):
            lattice_matrix = np.array([[1000.0, 0.0, 0.0], [0.0, 1000.0, 0.0], [0.0, 0.0, 1000.0]], dtype=float)
            pbc = np.array([0, 0, 0], dtype=int)
        else:
            raise ValueError("structure type not supported")
        if self.pbc is not None:
            pbc = np.array(self.pbc, dtype=int)
        r = float(cutoff)
        cart_coords = np.ascontiguousarray(np.array(structure.cart_coords), dtype=float)
        center_indices, neighbor_indices, images, distances = find_points_in_spheres(
            cart_coords, cart_coords, r=r, pbc=pbc, lattice=lattice_matrix, tol=numerical_tol
        )
        center_indices = center_indices.astype(np.int)
        neighbor_indices = neighbor_indices.astype(np.int)
        images = images.astype(np.int)
        distances = distances.astype(np.float32)
        exclude_self = (center_indices != neighbor_indices) | (distances > numerical_tol)
        return center_indices[exclude_self], neighbor_indices[exclude_self], images[exclude_self], distances[
            exclude_self]

    def get_graphs_within_cutoff_merge_sort(self, structure: StructureOrMolecule
                                            ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """For quick get bond distance"""
        cutoff = self.cutoff
        numerical_tol = self.numerical_tol
        center_indices, neighbor_indices, images, distances = self._within_cutoff(structure, cutoff,
                                                                                  numerical_tol)
        ele_numbers = np.array(structure.atomic_numbers)
        center_indices, center_prop, neighbor_indices, images, distances = self.refine(center_indices, neighbor_indices,
                                                                                       distances,
                                                                                       images,
                                                                                       center_prop=None,
                                                                                       ele_numbers=ele_numbers,
                                                                                       **self.refined_strategy_param,
                                                                                       )

        if len(center_indices.tolist()) == len(structure.species):
            return center_indices, center_prop, neighbor_indices, images, distances
        else:
            print("For {}, There is no neighbor in cutoff {} A, try with cutoff {} A for this structure.".format(
                str(structure.composition), cutoff, cutoff + 2)
            )
            center_indices, center_prop, neighbor_indices, images, distances = self.get_graphs_within_cutoff_merge_sort(
                structure)
            return center_indices, center_prop, neighbor_indices, images, distances

    def get_graphs_with_strategy_merge_sort(self, structure: StructureOrMolecule
                                            ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """For get bond distance with different strategy, for different nn_staagy could be rewrite."""
        # assert hasattr(self.nn_strategy, "cutoff")

        index1 = []
        index2 = []
        bonds = []
        center_prop = []

        image = []
        for n, neighbors in enumerate(self.nn_strategy.get_all_nn_info(structure)):
            index1.extend([n] * len(neighbors))
            index2i = [neighbor["site_index"] for neighbor in neighbors]
            bondsi = [neighbor["weight"] for neighbor in neighbors]
            imagei = [list(neighbor["image"]) + list(neighbor.get("add_image", (0,))) for neighbor in neighbors]
            center_propi = [neighbor.get("add_atom_prop", None) for neighbor in neighbors]
            index2.extend(index2i)
            image.extend(imagei)
            bonds.extend(bondsi)
            center_prop.append(center_propi)

        if None in center_prop[0] or [] in center_prop:
            center_prop = None
        else:
            center_prop = np.array(center_prop)

        ele_numbers = np.array(structure.atomic_numbers)
        center_indices, cen_prop, neighbor_indices, images, distances = self.refine(np.array(index1), np.array(index2),
                                                                                    np.array(bonds), np.array(image),
                                                                                    center_prop=center_prop,
                                                                                    ele_numbers=ele_numbers,
                                                                                    **self.refined_strategy_param)

        if len(center_indices.tolist()) == len(structure.species):
            return center_indices, cen_prop, neighbor_indices, images, distances
        else:

            if hasattr(self.nn_strategy, "min_bond_distance"):
                print("For {}, There is no neighbor in cutoff {} A, try with cutoff {} A for this structure.".format(
                    str(structure.composition), self.nn_strategy.min_bond_distance,
                    self.nn_strategy.min_bond_distance + 0.1)
                )
                self.nn_strategy.min_bond_distance += 0.1
            else:
                print("For {}, There is no neighbor in cutoff {} A, try with cutoff {} A for this structure.".format(
                    str(structure.composition), self.nn_strategy.cutoff, self.nn_strategy.cutoff + 2)
                )
            self.nn_strategy.cutoff += 2
            center_indices, center_prop, neighbor_indices, images, distances = self.get_graphs_with_strategy_merge_sort(
                structure)

            return center_indices, cen_prop, neighbor_indices, images, distances


#####################################################################################################################
after_treatment_func_map_des = {"universe_refine": universe_refine_des}

after_treatment_func_map_nn = {"universe_refine": universe_refine_nn}
# class
env_names = {"BaseNNGet": BaseNNGet, "BaseDesGet": BaseDesGet}
# local env method
env_method = {"BaseNNGet": NNDict, "BaseDesGet": DesDict, }
# after treatment
env_after_treatment_func_map = {"BaseNNGet": after_treatment_func_map_nn, "BaseDesGet": after_treatment_func_map_des, }
#####################################################################################################################



"""Abstract classes and utility operations for building graph representations and
preprocessing loaders."""

import warnings
from operator import itemgetter
from typing import Union, Dict, List, Any

import numpy as np
from mgetool.tool import parallelize, batch_parallelize
from pymatgen.analysis.local_env import NearNeighbors
from pymatgen.core import Structure

from featurebox.featurizers.atom.mapper import AtomPymatgenPropMap, AtomJsonMap, get_atom_fea_number, get_atom_fea_name
from featurebox.featurizers.base_transform import DummyConverter, BaseFeature, Converter
from featurebox.featurizers.bond.expander import BondGaussianConverter
from featurebox.featurizers.envir.environment import BaseNNGet, BaseDesGet, env_method, env_names
from featurebox.featurizers.envir.local_env import NNDict
from featurebox.utils.look_json import get_marked_class


def itemgetter_list(data_list: List, indices: List) -> tuple:
    """
    Get indices of data_list and return a tuple

    Args:
        data_list (list):  preprocessing list
        indices: (list) indices

    Returns:
        (tuple)
    """
    it = itemgetter(*indices)
    if np.size(indices) == 1:
        return it(data_list),
    return it(data_list)


class _StructureGraph(BaseFeature):
    """
    Preferential use of _StructureGraphFixedRadius.
    This is a base class for converting converting structure into graphs or model inputs.

    Methods to be implemented are follows:
        convert(self, structure)
            This is to convert a structure into a graph dictionary
    """

    def __init__(self, nn_strategy: Union[str, NearNeighbors] = "find_points_in_spheres",
                 bond_generator: [BaseNNGet, BaseDesGet, str] = None,
                 atom_converter: Converter = None,
                 bond_converter: Converter = None,
                 state_converter: Converter = None,
                 return_bonds: str = "all",
                 cutoff: float = 5.0,
                 **kwargs):
        """
        Args:
            nn_strategy (str): NearNeighbor strategy
                For bond_converter ="BaseNNGet": ["BrunnerNN_reciprocal", "BrunnerNN_real", "BrunnerNN_relative",
                "EconNN", "CrystalNN", "MinimumDistanceNNAll", "find_points_in_spheres"]
                For bond_converter ="BaseDesGet": ["ACSF","BehlerParrinello","EAD","EAMD","SOAP","SO3","SO4_Bispectrum","wACSF"]
                See Also:
                :class:`BaseNNGet` : :class:`CrystalNN` , :class:`MinimumDistanceNNAll` :class:`EconNN` .\n
                :class:`BaseDesGet` : :class:`ACSF` , :class:`SOAP` , :class:`EAMD`
            atom_converter (Converter): atom features converter.
                See Also:
                :class:`AtomTableMap` , :class:`AtomJsonMap` ,
                :class:`AtomPymatgenPropMap`, :class:`AtomTableMap`
            bond_converter (Converter): bond features converter, default=None.
            state_converter (Converter): state features converter.
                See Also:
                :class:`AtomPymatgenStrutureMap`
            bond_generator (_BaseEnvGet, str): bond features converter.
                The function of this, is to convert data format to a fixed format.
                "BaseNNGet" or "BaseDesGet" or defined BaseNNGet,BaseDesGet object, default "BaseNNGet".
                1, BaseDesGet or 2, BaseDesGet. or there name.
                if object offered, rather str, the nn_strategy would use the nn_strategy in Converter.
                See Also:
                :class:`BaseNNGet` , :class:`BaseDesGet`
            return_bonds:"all","bonds","bond_state"
                which bond property return. default "all".
                ``"bonds_state"`` : bond properties and ``"bonds"`` : atoms number near this center atom.
            cutoff: float
                Whether to use depends on the ``nn_strategy``.
            **kwargs:
        """

        super().__init__(**kwargs)
        self.return_bonds = return_bonds
        self.cutoff = cutoff

        if bond_generator is None:  # default use NNDict
            self.nn_strategy = get_marked_class(nn_strategy, NNDict)
            # there use the universal parameter, custom it please
            self.bond_generator = BaseNNGet(self.nn_strategy,
                                            numerical_tol=1e-8, pbc=None, cutoff=self.cutoff)
        elif isinstance(bond_generator, str):  # new add "BaseDesGet"
            self.nn_strategy = get_marked_class(nn_strategy, env_method[bond_generator])
            # there use the universal parameter, custom it please
            self.bond_generator = env_names[bond_generator](self.nn_strategy, self.cutoff,
                                                            numerical_tol=1e-8, pbc=None, cutoff=self.cutoff)
        else:  # defined BaseDesGet or BaseNNGet
            self.bond_generator = bond_generator
            self.nn_strategy = self.bond_generator.nn_strategy
        self.atom_converter = atom_converter or self._get_dummy_converter()
        self.bond_converter = bond_converter or self._get_dummy_converter()
        self.state_converter = state_converter or self._get_dummy_converter()
        self.graph_data_name = ["atom", "bond", "state", 'atom_nbr_idx']
        self.cutoff = cutoff

    def __add__(self, other):
        raise TypeError("There is no add.")

    def convert(self, structure: Structure, state_attributes: List = None) -> Dict:
        """
        Take a pymatgen structure and convert it to a index-type graph representation
        The graph will have node, distance, index1, index2, where node is a vector of Z number
        of atoms in the structure, index1 and index2 mark the atom indices forming the bond and separated by
        distance.

        For state attributes, you can set structure.state = [[xx, xx]] beforehand or the algorithm would
        take default [[0, 0]]

        Args:
            state_attributes: (list) state attributes
            structure: (pymatgen structure)
            (dictionary)
        """
        if state_attributes is not None:
            state_attributes = np.array(state_attributes)
        else:
            state_attributes = np.array([0.0, 0.0], dtype="float32")
        if isinstance(self.state_converter, DummyConverter):
            pass
        else:
            state_attributes = np.concatenate(
                (state_attributes, np.array(self.state_converter.convert(structure)).ravel()))
        center_indices, center_prop, atom_nbr_idx, bond_states, bonds = self.get_bond_fea(structure)
        if self.return_bonds == "all":
            bondss = np.concatenate((bond_states, bonds), axis=-1) if bonds is not None else bond_states
        elif self.return_bonds == "bonds":
            if bonds is not None:
                bondss = bonds
            else:
                raise TypeError
        elif self.return_bonds == "bonds_state":
            bondss = bond_states
        else:
            raise NotImplementedError()

        atoms = self.get_atom_fea(structure)
        atoms = [atoms[round(i)] for i in center_indices]
        atoms = self.atom_converter.convert(atoms)
        bonds = self.bond_converter.convert(bondss)

        if center_prop.shape[1] > 1:
            atoms = np.concatenate((np.array(atoms), center_prop), axis=1)

        return {"atom": atoms, "bond": bonds, "state": state_attributes, "atom_nbr_idx": atom_nbr_idx}

    @staticmethod
    def get_atom_fea(structure: Structure) -> List[Any]:
        """
        Get atom features from structure, may be overwritten for you self
        # todo ding yi zi ji de yuan zi

        Args:
            structure: (Pymatgen.Structure) pymatgen structure

        Returns:
            List of atomic numbers
        """
        return get_atom_fea_number(structure)

    def get_bond_fea(self, structure: Structure):
        """
        Get atom features from structure, may be overwritten
        # todo ding yi zi ji de jian.

        Args:
            structure: (Pymatgen.Structure) pymatgen structure
        """
        return self.bond_generator.convert(structure)

    def __call__(self, structure: Structure, *args, **kwargs) -> Dict:
        """
        Directly apply the converter to structure, alias to convert.

        Args:
            structure (Structure): input structure

        Returns (dict): graph dictionary

        """
        return self.convert(structure, *args, **kwargs)

    @staticmethod
    def _get_dummy_converter() -> DummyConverter:
        return DummyConverter()

    def _transform(self, structures: List[Structure], state_attributes: List = None):
        """

        Args:
            structures:list
                preprocessing of samples need to transform to Graph.
            state_attributes
                preprocessing of samples need to add to Graph.

        Returns:
            list of graphs: List of dict

        """

        if state_attributes is None:
            state_attributes = [None] * len(structures)
        iterables = zip(structures, state_attributes)

        if not self.batch_calculate:
            rets = parallelize(self.n_jobs, self._wrapper, iterables, tq=True)

            ret, self.support_ = zip(*rets)

        else:
            rets = batch_parallelize(self.n_jobs, self._wrapper, iterables, respective=False,
                                     tq=True, batch_size=self.batch_size)

            ret, self.support_ = zip(*rets)
        return ret

    def get_flat_data(self, graphs: List[Dict]) -> tuple:
        """
        Expand the graph dictionary to form a list of features and targets tensors.
        This is useful when the model is trained on assembled graphs on the fly.
        Aim to get input for GraphGenerator.

        Args:
            graphs: (list of dictionary) list of graph dictionary for each structure

        Returns:
            tuple(node_features, edges_features, stats_values, atom_nbr_idx , ***)
        """
        output = []  # Will be a list of arrays

        # Convert the graphs to matrices
        for n, fi in enumerate(self.graph_data_name):
            output.append([np.array(gi[fi]) if isinstance(gi, dict) else np.array(gi[n]) for gi in graphs])

        return tuple(output)

    def transform(self, structures: List[Structure], state_attributes: List = None):
        """

        Parameters
        ----------
        structures:list
            preprocessing of samples need to transform to Graph.
        state_attributes
            preprocessing of samples need to add to Graph.

        Returns
        -------
        list of graphs: List of preprocessing
        """
        return self.get_flat_data(self._transform(structures, state_attributes))


class _StructureGraphFixedRadius(_StructureGraph):
    """
    Preferential use of _StructureGraphFixedRadius.
    A phase radius is recommended: cutoff.
    """

    def __init__(
            self,
            nn_strategy: Union[str, NearNeighbors] = "find_points_in_spheres",
            atom_converter: Converter = None,
            bond_converter: Converter = None,
            state_converter: Converter = None,
            cutoff: float = 7.0,
            **kwargs
    ):
        if cutoff is None:
            warnings.warn(UserWarning("A phase radius is recommended: cutoff."))
        super().__init__(
            nn_strategy=nn_strategy, atom_converter=atom_converter, bond_converter=bond_converter,
            state_converter=state_converter, cutoff=cutoff, **kwargs)


class CrystalGraph(_StructureGraphFixedRadius):
    """
    CrystalGraph.

    Examples
    --------
    >>> cg1 = CrystalGraph()
    >>> d = cg1(structure)
    >>> d = cg1(structure, state_attributes=np.array([2,3.0]))
    >>> d = cg1.convert(structure, state_attributes=np.array([2,3.0]))
    >>> ds = cg1.fit_transform(structures)

    """

    def __init__(
            self,
            nn_strategy: Union[str, NearNeighbors] = "find_points_in_spheres",
            atom_converter: Converter = None,
            bond_converter: Converter = None,
            state_converter: Converter = None,
            cutoff: float = 7.0,
            **kwargs
    ):
        super().__init__(
            nn_strategy=nn_strategy, atom_converter=atom_converter, bond_converter=bond_converter,
            state_converter=state_converter, cutoff=cutoff, **kwargs)


class CrystalGraphWithBondTypes(_StructureGraph):
    """
    Overwrite the bond attributes with bond types, defined simply by
    the metallicity of the atoms forming the bond. Three types of
    scenario is considered, nonmetal-nonmetal (type 0), metal-nonmetal (type 1), and
    metal-metal (type 2)

    Examples
    --------
    >>> cg1 = CrystalGraphWithBondTypes()
    >>> d = cg1(structure)
    >>> d = cg1(structure, state_attributes=np.array([2,3.0]))
    >>> d = cg1.convert(structure, state_attributes=np.array([2,3.0]))
    >>> ds = cg1.fit_transform(structures)

    """

    def __init__(
            self,
            nn_strategy: Union[str, NearNeighbors] = "VoronoiNN",
            atom_converter: Converter = None,
            bond_converter: Converter = None,
            state_converter: Converter = None,
            return_bonds="bonds",
            **kwargs
    ):
        if bond_converter is None:
            bond_converter = AtomPymatgenPropMap("is_metal", func=None, search_tp="number")

        super().__init__(nn_strategy=nn_strategy, atom_converter=atom_converter, bond_converter=bond_converter,
                         state_converter=state_converter, return_bonds=return_bonds,
                         **kwargs)


class CrystalGraphDisordered(_StructureGraphFixedRadius):
    """
    Enable disordered site predictions

    Examples
    --------
    >>> cg1 = CrystalGraphDisordered()
    >>> d = cg1(structure)
    >>> d = cg1(structure, state_attributes=np.array([2,3.0]))
    >>> d = cg1.convert(structure, state_attributes=np.array([2,3.0]))
    >>> ds = cg1.fit_transform(structures)


    """

    def __init__(
            self,
            nn_strategy: Union[str, NearNeighbors] = "VoronoiNN",
            atom_converter: Converter = AtomJsonMap(),
            bond_converter: Converter = None,
            state_converter: Converter = None,
            cutoff: float = 7.0,
            **kwargs
    ):
        self.cutoff = cutoff
        super().__init__(
            nn_strategy=nn_strategy, atom_converter=atom_converter, bond_converter=bond_converter, cutoff=self.cutoff,
            state_converter=state_converter,
            **kwargs)

    @staticmethod
    def get_atom_fea(structure) -> List[dict]:
        """
        For a structure return the list of dictionary for the site occupancy
        for example, Fe0.5Ni0.5 site will be returned as {"Fe": 0.5, "Ni": 0.5}

        Args:
            structure (Structure): pymatgen Structure with potential site disorder

        Returns:
            a list of site fraction description
        """
        return get_atom_fea_name(structure)


class SimpleMolGraph(_StructureGraphFixedRadius):
    """
    Default using all atom pairs as bonds. The distance between atoms are used
    as bond features. By default the distance is expanded using a Gaussian
    expansion with centers at np.linspace(0, 4, 20) and width of 0.5

    Examples
    --------
    >>> cg1 = SimpleMolGraph()
    >>> d = cg1(structure)
    >>> d = cg1(structure, state_attributes=np.array([2,3.0]))
    >>> d = cg1.convert(structure, state_attributes=np.array([2,3.0]))
    >>> ds = cg1.fit_transform(structures)
    """

    def __init__(
            self,
            nn_strategy: Union[str, NearNeighbors] = "AllAtomPairs",
            atom_converter: Converter = None,
            bond_converter: Converter = None,
            state_converter: Converter = None,
            **kwargs):

        if bond_converter is None:
            bond_converter = BondGaussianConverter(np.linspace(0, 4, 20), 0.5)
        super().__init__(nn_strategy=nn_strategy, atom_converter=atom_converter, bond_converter=bond_converter,
                         state_converter=state_converter,
                         **kwargs)



#!/usr/bin/python3.7
# -*- coding: utf-8 -*-

# @Time   : 2019/7/27 16:57
# @Author : Administrator
# @Software: PyCharm
# @License: BSD 3-Clause

"""
This is script, copy for using, rather than call.**
"""

import warnings
from functools import partial

import numpy as np
from sklearn import kernel_ridge, gaussian_process, neighbors
from sklearn.ensemble import AdaBoostClassifier, AdaBoostRegressor, RandomForestClassifier, GradientBoostingRegressor, \
    RandomForestRegressor, GradientBoostingClassifier
from sklearn.gaussian_process.kernels import RBF, Matern
from sklearn.linear_model import LogisticRegression, BayesianRidge, SGDRegressor, Lasso, ElasticNet, Perceptron, \
    SGDClassifier, PassiveAggressiveRegressor
from sklearn.metrics import get_scorer
from sklearn.model_selection import StratifiedKFold, GridSearchCV, cross_val_score
from sklearn.svm import SVR, SVC
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor

warnings.filterwarnings("ignore")

kernel = 1.0 * RBF(1.0)
kernel2 = Matern(nu=0.5)
kernel3 = Matern(nu=1.5)
kernel4 = Matern(nu=2.5)
kernel5 = Matern(nu=0.5, length_scale=0.8)
kernel6 = Matern(nu=1.5, length_scale=0.8)
kernel7 = Matern(nu=2.5, length_scale=0.8)
kernel8 = 2 * Matern(nu=0.5)
kernel9 = 2 * Matern(nu=1.5)
kernel10 = 2 * Matern(nu=2.5)

ker = [kernel, kernel2, kernel3, kernel4, kernel5, kernel6, kernel7, kernel8, kernel9, kernel10]


def dict_method_clf():
    """many clf method."""
    dict_method = {}
    # 1st part
    """4KNC"""
    me4 = neighbors.KNeighborsClassifier(n_neighbors=5)
    cv4 = StratifiedKFold(5, shuffle=False, random_state=0)
    scoring4 = 'balanced_accuracy'

    param_grid4 = [{'n_neighbors': [3, 4, 5, 6, 7], "weight": ['uniform', "distance"], "leaf_size=30": [10, 20, 30],
                    'metric': ['seuclidean', "manhattan"]}, ]

    dict_method.update({"KNC-set": [me4, cv4, scoring4, param_grid4]})

    """1SVC"""
    me1 = SVC(C=1.0, kernel='rbf', degree=3, gamma='auto_deprecated',
              coef0=0.0, shrinking=True, probability=False,
              tol=1e-3, cache_size=200, class_weight='balanced',
              verbose=False, max_iter=-1, decision_function_shape='ovr',
              random_state=None)
    cv1 = StratifiedKFold(5, shuffle=False)
    scoring1 = 'accuracy'

    param_grid1 = [{'C': [1.0e8, 1.0e6, 10000, 100, 50, 10, 5, 2.5, 1, 0.5, 0.1, 0.01], 'kernel': ker}]

    dict_method.update({'SVC-set': [me1, cv1, scoring1, param_grid1]})

    """5GPC"""
    me5 = gaussian_process.GaussianProcessClassifier(kernel=kernel)
    cv5 = StratifiedKFold(5, shuffle=False)
    scoring5 = 'balanced_accuracy'
    param_grid5 = [{"kernel": ker}, ]

    dict_method.update({'GPC-set': [me5, cv5, scoring5, param_grid5]})

    # 2nd part
    '''TreeC'''
    me6 = DecisionTreeClassifier(
        criterion='gini', splitter='best', max_depth=None, min_samples_split=2, min_samples_leaf=1,
        min_weight_fraction_leaf=0.0, max_features=None, random_state=None, max_leaf_nodes=None,
        min_impurity_decrease=0.0, min_impurity_split=None, class_weight="balanced", presort=False)
    cv6 = StratifiedKFold(5, shuffle=False)
    scoring6 = 'accuracy'
    param_grid6 = [{'max_depth': [3, 4, 5, 6, 7, 8, 9, 10], 'min_samples_split': [2, 3, 4]}]
    dict_method.update({'TreeC-em': [me6, cv6, scoring6, param_grid6]})

    '''GBC'''
    me7 = GradientBoostingClassifier(
        loss='deviance', learning_rate=0.1, n_estimators=100,
        subsample=1.0, criterion='friedman_mse', min_samples_split=2,
        min_samples_leaf=1, min_weight_fraction_leaf=0.,
        max_depth=3, min_impurity_decrease=0.,
        min_impurity_split=None, init=None,
        random_state=None, max_features=None, verbose=0,
        max_leaf_nodes=None, warm_start=False,
        presort='auto')
    cv7 = StratifiedKFold(5, shuffle=False)
    scoring7 = 'balanced_accuracy'
    param_grid7 = [{'n_estimators': [50, 100, 200, 500], 'max_depth': [3, 4, 5, 6, 7, 8, 9, 10],
                    'min_samples_split': [2, 3, 4], 'learning_rate': [0.1, 0.05]}]
    dict_method.update({'GBC-em': [me7, cv7, scoring7, param_grid7]})

    '''RFC'''
    me8 = RandomForestClassifier(n_estimators=100, criterion="gini", max_depth=None,
                                 min_samples_split=2, min_samples_leaf=1, min_weight_fraction_leaf=0.,
                                 max_features="auto", max_leaf_nodes=None, min_impurity_decrease=0.,
                                 min_impurity_split=None, bootstrap=True, oob_score=False,
                                 random_state=None, verbose=0, warm_start=False,
                                 class_weight="balanced")
    cv8 = StratifiedKFold(5, shuffle=False)
    scoring8 = 'accuracy'
    param_grid8 = [{'n_estimators': [50, 100, 200, 500], 'max_depth': [3, 4, 5, 6, 7, 8, 9, 10],
                    'min_samples_split': [2, 3, 4], 'learning_rate': [0.1, 0.05]}]
    dict_method.update({"RFC-em": [me8, cv8, scoring8, param_grid8]})

    "AdaBC"
    dt = DecisionTreeRegressor(criterion="mse", splitter="best", max_features=None, max_depth=5, min_samples_split=4)
    me9 = AdaBoostClassifier(dt, n_estimators=100, learning_rate=1., algorithm='SAMME.R',
                             random_state=0)
    cv9 = StratifiedKFold(5, shuffle=False)
    scoring9 = 'accuracy'
    param_grid9 = [{'n_estimators': [50, 100, 200, 500], 'max_depth': [3, 4, 5, 6, 7, 8, 9, 10],
                    'min_samples_split': [2, 3, 4], 'learning_rate': [0.1, 0.05]}]
    dict_method.update({"AdaBC-em": [me9, cv9, scoring9, param_grid9]})

    # 3nd

    "Per"
    me14 = Perceptron(penalty="l1", alpha=0.0001, fit_intercept=True, max_iter=None, tol=None,
                      shuffle=True, verbose=0, eta0=1.0, random_state=0,
                      class_weight=None, warm_start=False, n_iter=None)
    cv14 = StratifiedKFold(5, shuffle=False)
    scoring14 = 'accuracy'
    param_grid14 = [{'alpha': [0.0001, 0.001, 0.01]}, ]
    dict_method.update({"Per-L1": [me14, cv14, scoring14, param_grid14]})

    """LogRL1"""
    me15 = LogisticRegression(penalty='l1', solver='liblinear', dual=False, tol=1e-3, C=1.0, fit_intercept=True,
                              intercept_scaling=1, class_weight='balanced', random_state=0)
    cv15 = StratifiedKFold(5, shuffle=False)
    scoring15 = 'accuracy'
    param_grid15 = [{'C': [0.1, 0.2, 0.3, 0.4, 0.5, 1, 2], 'penalty': ["l1", "l2"]}, ]
    dict_method.update({"LogR-L1": [me15, cv15, scoring15, param_grid15]})

    """3SGDCL2"""
    me3 = SGDClassifier(loss='hinge', penalty='l2', alpha=0.0001, l1_ratio=0.15,
                        fit_intercept=True, max_iter=None, tol=None, shuffle=True,
                        verbose=0, epsilon=0.1, random_state=0,
                        learning_rate='optimal', eta0=0.0, power_t=0.5,
                        class_weight="balanced", warm_start=False, average=False, n_iter=None)
    cv3 = StratifiedKFold(5, shuffle=False)
    scoring3 = 'accuracy'

    param_grid3 = [{'alpha': [100, 10, 1, 0.1, 0.01, 0.001, 0.0001, 1e-05], 'loss': ['squared_loss', "huber"],
                    "penalty": ["l1", "l2"]}, ]

    dict_method.update({"SGDC-set": [me3, cv3, scoring3, param_grid3]})

    return dict_method


def dict_method_reg():
    """many reg method."""
    dict_method = {}
    # 1st part

    """4KNR"""
    me4 = neighbors.KNeighborsRegressor(n_neighbors=5, weights='uniform', algorithm='auto', leaf_size=30, p=2,
                                        metric='minkowski')
    cv4 = 5
    scoring4 = 'r2'
    param_grid4 = [{'n_neighbors': [3, 4, 5, 6, 7], "weights": ['uniform', "distance"], "leaf_size": [10, 20, 30]
                    }]
    dict_method.update({"KNR-set": [me4, cv4, scoring4, param_grid4]})

    """1SVR"""
    me1 = SVR(kernel='rbf', gamma='auto', degree=3, tol=1e-3, epsilon=0.1, shrinking=False, max_iter=2000)
    cv1 = 5
    scoring1 = 'r2'
    param_grid1 = [{'C': [10000, 100, 50, 10, 5, 2.5, 1, 0.5, 0.1, 0.01], 'kernel': ker}]
    dict_method.update({"SVR-set": [me1, cv1, scoring1, param_grid1]})

    """5kernelridge"""
    me5 = kernel_ridge.KernelRidge(alpha=1, gamma="scale", degree=3, coef0=1, kernel_params=None)
    cv5 = 5
    scoring5 = 'r2'
    param_grid5 = [{'alpha': [100, 50, 10, 5, 2.5, 1, 0.5, 0.1, 0.01, 0.001, 1e-4, 1e-5], 'kernel': ker}]
    dict_method.update({'KRR-set': [me5, cv5, scoring5, param_grid5]})

    """6GPR"""
    me6 = gaussian_process.GaussianProcessRegressor(kernel=kernel, alpha=1e-10, optimizer='fmin_l_bfgs_b',
                                                    n_restarts_optimizer=0,
                                                    normalize_y=False, copy_X_train=True, random_state=0)
    cv6 = 5
    scoring6 = 'r2'
    param_grid6 = [{'alpha': [1e-3, 1e-2], 'kernel': ker}]
    dict_method.update({"GPR-set": [me6, cv6, scoring6, param_grid6]})

    # 2nd part

    """6RFR"""
    me7 = RandomForestRegressor(n_estimators=100, max_depth=None, min_samples_split=2, min_samples_leaf=1,
                                min_weight_fraction_leaf=0.0, max_leaf_nodes=None, min_impurity_decrease=0.0,
                                min_impurity_split=None, bootstrap=True, oob_score=False,
                                random_state=None, verbose=0, warm_start=False)
    cv7 = 5
    scoring7 = 'r2'
    param_grid7 = [{'max_depth': [3, 4, 5], }]
    dict_method.update({"RFR-em": [me7, cv7, scoring7, param_grid7]})

    """7GBR"""
    me8 = GradientBoostingRegressor(loss='ls', learning_rate=0.1, n_estimators=100,
                                    subsample=1.0, criterion='friedman_mse', min_samples_split=2,
                                    min_samples_leaf=1, min_weight_fraction_leaf=0.,
                                    max_depth=3, min_impurity_decrease=0.,
                                    min_impurity_split=None, init=None, random_state=None,
                                    max_features=None, alpha=0.9, verbose=0, max_leaf_nodes=None,
                                    warm_start=False, presort='auto')
    cv8 = 5
    scoring8 = 'r2'
    param_grid8 = [{'max_depth': [3, 4, 5, 6], 'min_samples_split': [2, 3],
                    'learning_rate': [0.1, 0.05]}]
    dict_method.update({'GBR-em': [me8, cv8, scoring8, param_grid8]})

    "AdaBR"
    dt = DecisionTreeRegressor(criterion="mse", splitter="best", max_features=None, max_depth=5, min_samples_split=4)
    me9 = AdaBoostRegressor(dt, n_estimators=200, learning_rate=0.05, loss='linear', random_state=0)
    cv9 = 5
    scoring9 = 'explained_variance'
    param_grid9 = [{'n_estimators': [100, 200], 'learning_rate': [0.1, 0.05]}]
    dict_method.update({"AdaBR-em": [me9, cv9, scoring9, param_grid9]})

    '''DTR'''
    me10 = DecisionTreeRegressor(
        criterion="mse",
        splitter="best",
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_weight_fraction_leaf=0.,
        max_features=None,
        random_state=0,
        max_leaf_nodes=None,
        min_impurity_decrease=0.,
        min_impurity_split=None,
        presort=False)
    cv10 = 5
    scoring10 = 'r2'
    param_grid10 = [
        {'max_depth': [2, 3, 4, 5, 6, 7, 8], "min_samples_split": [2, 3, 4], "min_samples_leaf": [1, 2]}]
    dict_method.update({'DTR-em': [me10, cv10, scoring10, param_grid10]})

    'ElasticNet'
    me11 = ElasticNet(alpha=1.0, l1_ratio=0.7, fit_intercept=True, normalize=False, precompute=False, max_iter=1000,
                      copy_X=True, tol=0.0001, warm_start=False, positive=False, random_state=None)

    cv11 = 5
    scoring11 = 'r2'
    param_grid11 = [{'alpha': [0.0001, 0.001, 0.01, 0.1, 1, 10], 'l1_ratio': [0.3, 0.5, 0.8]}]
    dict_method.update({"EN-L1": [me11, cv11, scoring11, param_grid11]})

    'Lasso'
    me12 = Lasso(alpha=1.0, fit_intercept=True, normalize=False, precompute=False, copy_X=True, max_iter=1000,
                 tol=0.001,
                 warm_start=False, positive=False, random_state=None, )

    cv12 = 5
    scoring12 = 'r2'
    param_grid12 = [{'alpha': [0.0001, 0.0005, 0.001, 0.005, 0.01, 0.05, 0.1, 0.5, 1, 10, 100, 1000]}, ]
    dict_method.update({"LASSO-L1": [me12, cv12, scoring12, param_grid12]})

    """2BayesianRidge"""
    me2 = BayesianRidge(alpha_1=1e-06, alpha_2=1e-06, compute_score=False,
                        copy_X=True, fit_intercept=True, lambda_1=1e-06, lambda_2=1e-06,
                        n_iter=300, normalize=False, tol=0.01, verbose=False)
    cv2 = 5
    scoring2 = 'r2'
    param_grid2 = [{'alpha_1': [1e-07, 1e-06, 1e-05], 'alpha_2': [1e-07, 1e-06, 1e-05]}]
    dict_method.update({'BRR-L1': [me2, cv2, scoring2, param_grid2]})

    """3SGDRL2"""
    me3 = SGDRegressor(alpha=0.0001, average=False,
                       epsilon=0.1, eta0=0.01, fit_intercept=True, l1_ratio=0.15,
                       learning_rate='invscaling', loss='squared_loss', max_iter=1000,
                       penalty='l2', power_t=0.25,
                       random_state=0, shuffle=True, tol=0.01,
                       verbose=0, warm_start=False)
    cv3 = 5
    scoring3 = 'r2'
    param_grid3 = [{'alpha': [100, 10, 1, 0.1, 0.01, 0.001, 0.0001, 1e-05], 'loss': ['squared_loss', "huber"],
                    "penalty": ["l1", "l2"]}]
    dict_method.update({'SGDR-L1': [me3, cv3, scoring3, param_grid3]})

    """PassiveAggressiveRegressor"""
    me14 = PassiveAggressiveRegressor(C=1.0, fit_intercept=True, max_iter=1000, tol=0.001, early_stopping=False,
                                      validation_fraction=0.1, n_iter_no_change=5, shuffle=True, verbose=0,
                                      loss='epsilon_insensitive', epsilon=0.1, random_state=None,
                                      warm_start=False, average=False)
    cv14 = 5
    scoring14 = 'r2'
    param_grid14 = [{'C': [1.0e8, 1.0e6, 10000, 100, 50, 10, 5, 2.5, 1, 0.5, 0.1, 0.01]}]
    dict_method.update({'PAR-L1': [me14, cv14, scoring14, param_grid14]})

    return dict_method


def dict_me(me="clf"):
    if me == "clf":
        dict_method_ = dict_method_clf()
    else:
        dict_method_ = dict_method_reg()
    return dict_method_


def method_pack(method_all, me="reg", scoreing=None, gd=True, cv=10):
    """return cv or gd."""
    if not method_all:
        method_all = ['KNR-set', 'SVR-set', "KRR-set", "GPR-set",
                      "RFR-em", "AdaBR-em", "DTR-em",
                      "LASSO-L1", "BRR-L1", "SGDR-L1", "PAR-L1"]
    dict_method = dict_me(me=me)

    print(dict_method.keys())
    if gd:
        estimator = []
        for method_i in method_all:
            me2, cv2, scoring2, param_grid2 = dict_method[method_i]
            if me == "clf":
                scoring2 = scoreing if scoreing else 'balanced_accuracy'
            if me == "reg":
                scoring2 = scoreing if scoreing else 'r2'
            cv2 = cv if cv else cv2
            gd2 = GridSearchCV(me2, cv=cv2, param_grid=param_grid2, scoring=scoring2, n_jobs=1)
            estimator.append(gd2)
        return estimator
    else:
        estimator = []
        for method_i in method_all:
            me2, cv2, scoring2, param_grid2 = dict_method[method_i]
            if me == "clf":
                scoring2 = scoreing if scoreing else 'balanced_accuracy'
            if me == "reg":
                scoring2 = scoreing if scoreing else 'r2'
            cv2 = cv if cv else cv2
            gd2 = partial(cross_val_score, estimator=me2, cv=cv2, scoring=scoring2)
            # gd2 = cross_val_score(me2, cv=cv2, scoring=scoring2)
            estimator.append(gd2)
        return estimator


def cv_predict(x, y, s_estimator, kf):
    y_test_predict_all = []
    for train, test in kf:
        X_train, X_test, y_train, y_test = x[train], x[test], y[train], y[test]
        s_estimator.fit(X_train, y_train)
        y_test_predict = s_estimator.predict(X_test)
        y_test_predict_all.append(y_test_predict)

    test_index = [i[1] for i in kf]
    y_test_true_all = [y[_] for _ in test_index]

    return y_test_true_all, y_test_predict_all


def pack_score(y_test_true_all, y_test_predict_all, scoring):
    if isinstance(y_test_true_all, np.ndarray) and isinstance(y_test_predict_all, np.ndarray):
        y_test_true_all = [y_test_true_all, ]
        y_test_predict_all = [y_test_predict_all, ]
    if scoring == "rmse":
        scoring2 = 'neg_mean_squared_error'
    else:
        scoring2 = scoring

    scorer = get_scorer(scoring2)

    scorer_func = scorer._score_func

    score = [scorer_func(i, j) for i, j in zip(y_test_true_all, y_test_predict_all)]

    if scoring == "rmse":
        score = np.sqrt(score)
    score_mean = np.mean(score)
    return score_mean

# def my_score(gd_method, train_X, test_X, train_Y, test_Y):
#     # train_X, test_X = my_pca(train_X, test_X, 0.9)
#
#     # s_X = preprocessing.StandardScaler()
#     # train_X = s_X.fit_transform(train_X)
#     # test_X = s_X.transform(test_X)
#
#     # train_X,train_Y,= utils.shuffle(train_X,train_Y, random_state=3)
#
#     grid = gd_method
#     n_splits = 5
#     kf = KFold(n_splits=n_splits, shuffle=False)
#     grid.cv = kf
#     grid.fit(train_X, train_Y)
#
#     print("最好的参数：", grid.best_params_)
#     print("最好的模型：", grid.best_estimator_)
#     print("\ngrid.best_score:\n", grid.best_score_)
#
#     metrics_method1 = "rmse"
#     metrics_method2 = "r2"
#
#     # cv_train
#     kf = KFold(n_splits=n_splits, shuffle=False)
#     kf = list(kf.split(train_X))
#     y_train_true_all, y_train_predict_all = cv_predict(train_X, train_Y, grid.best_estimator_, kf)
#
#     cv_pre_train_y = np.concatenate([i.ravel() for i in y_train_predict_all]).T
#     score1 = pack_score(y_train_true_all, y_train_predict_all, metrics_method1)
#     score2 = pack_score(y_train_true_all, y_train_predict_all, metrics_method2)
#     print("train_X's cv score %s" % score1, "train_X's cv score %s" % score2)
#
#     # all_train
#     grid = grid.best_estimator_
#     # grid.fit(train_X,train_Y)
#     pre_train_y = grid.predict(train_X)
#     score3 = pack_score(train_Y, pre_train_y, metrics_method1)
#     score4 = pack_score(train_Y, pre_train_y, metrics_method2)
#     print("train_X's score %s " % score3, "train_X's score %s" % score4)
#     # test
#     pre_test_y = grid.predict(test_X)
#     score5 = pack_score(test_Y, pre_test_y, metrics_method1)
#     score6 = pack_score(test_Y, pre_test_y, metrics_method2)
#     print("test_X's score %s" % score5, "test_X's score %s" % score6)
#
#     return cv_pre_train_y, pre_train_y, pre_test_y

# if __name__ == "__main__":
#     import pandas as pd
#
#     com_data_raw = pd.read_csv('wxxwcx.csv')
#
#     com_data = com_data_raw
#
#     select_X5 = ['O-M-outer', 'a Lattice parameter', 'M Electron affinity mean',
#     'M first ionization potential differ',
#                  'valence X atom']
#     target_y = 'Hydrogen adsorption energy'
#     x = com_data[select_X5].values
#     y = com_data[target_y].values
#     method = ["SVR-set", 'KRR-set', "GPR-set", "RFR-em", 'GBR-em', "AdaBR-em", 'TreeC-em']
#     gd_method = method_pack(method, me="reg", gd=True)
#     train_X, test_X, train_Y, test_Y = x[42:, :], x[:42, :], y[42:], y[:42]
#     cv_pre_train_y, pre_train_y, pre_test_y = my_score(gd_method[5], train_X, test_X, train_Y, test_Y)



import unittest

from featurebox.featurizers.atom.mapper import AtomTableMap, AtomJsonMap
from featurebox.featurizers.state.statistics import WeightedAverage, WeightedSum


class MyTestCase(unittest.TestCase):

    def test_WeightedAverage(self):
        data_map = AtomTableMap(search_tp="name", n_jobs=1)
        wa = WeightedAverage(data_map, n_jobs=1)
        x3 = [{"H": 2, }, {"Pd": 1}]
        x4 = wa.fit_transform(x3)
        print(x4)

    def test_WeightedSum(self):
        data_map = AtomTableMap(search_tp="name", n_jobs=1)
        wa = WeightedSum(data_map, n_jobs=1)
        x3 = [{"H": 2, "Pd": 1}]
        x4 = wa.fit_transform(x3)
        print(x4)

    def test_WeightedSum2(self):
        data_map = AtomJsonMap(search_tp="name", n_jobs=1)
        wa = WeightedSum(data_map, n_jobs=1)
        x3 = [{"H": 2, "Pd": 1}, {"He": 1, "Al": 4}]
        x4 = wa.fit_transform(x3)
        print(x4)


if __name__ == '__main__':
    unittest.main()



import torch
from torch.nn import Module

from featurebox.featurizers.generator import MGEDataLoader


class BaseLearning:
    def __init__(self, model: Module, train_loader: MGEDataLoader, test_loader: MGEDataLoader = None, device="cpu",
                 opt=None, clf=False, loss_method=None):

        self.train_loader = train_loader
        self.test_loader = test_loader

        device = torch.device(device)
        self.train_loader.to_cuda(device)
        if self.test_loader is not None:
            self.test_loader.to_cuda(device)
        self.device = device
        self.model = model
        self.model.to(device)
        self.opt = opt
        self.clf = clf
        self.loss_method = loss_method
        self.train_batch_number = len(self.train_loader.loader)
        self.test_batch_number = len(self.test_loader.loader) if self.test_loader is not None else 0

    def method(self, opt=None, clf=False, loss_method=None, learning_rate=1e-3):
        """ "sum" is recommended, for batch training"""
        opt = opt if opt else self.opt
        if opt is None:
            self.opt = torch.optim.Adam(self.model.parameters(), lr=learning_rate, weight_decay=0.01)
            # L2 regularization
        else:
            self.opt = opt(self.model.parameters(), lr=learning_rate, weight_decay=0.01)

        if loss_method is None:
            if clf is True:
                self.loss_fn = torch.nn.CrossEntropyLoss(reduction='sum')
            elif clf == "multi_label":
                self.loss_fn = torch.nn.L1Loss(reduction='sum')
                # 主要是用来判定实际的输出与期望的输出的接近程度 MAE: 1/N |y_pred-y|
            else:
                self.loss_fn = torch.nn.MSELoss(reduction='sum')

        else:
            self.loss_fn = loss_method(reduction="sum")

    def run_train(self, epochs=30, threshold=230):
        #############defination###########################

        if self.opt is None:
            self.method()

        self.model.to(self.device)
        self.model.train()

        train_loader = self.train_loader

        ##############tarin###########################
        for epochi in range(epochs):

            train_loader.reset()
            loss = 0
            for m, (batch_x, batch_y) in enumerate(train_loader):
                self.opt.zero_grad()
                y_pred = self.model(*batch_x)
                lossi = self.loss_fn(y_pred, batch_y)
                loss += lossi.item()
                lossi.backward()
                self.opt.step()
            if loss <= threshold:
                break
            print("Train {} epochi loss:".format(epochi), loss)  #

    def run_test(self, force=True):
        ##############test###########################
        self.model.eval()

        if self.test_loader is None:
            if force:
                print("There is no test data loader. try with train data")
                test_loader = self.train_loader
            else:
                raise TypeError("There is no test loader!")
        else:
            test_loader = self.test_loader

        loss = 0.0
        for m, (batch_x, batch_y) in enumerate(test_loader):
            y_pred = self.model(*batch_x)
            lossi = self.loss_fn(y_pred, batch_y)
            loss += lossi.item()

        print("Test loss", loss)  #



"""Base"""
from multiprocessing import cpu_count
from typing import List, Tuple, Iterable, Any

import numpy as np
import pandas as pd
from mgetool.tool import parallelize
from monty.json import MSONable


class BaseFeature(MSONable):
    """
    **Using a BaseFeature Class**

    Abstract class to calculate features from ``MSONable``.
    That means you can embed this feature directly into ``BaseFeature`` class implement.
    ::

        class MatFeature(BaseFeature):
            def convert(spath, *x):
                ...

    ``BaseFeature`` implement :class:`sklearn.base.BaseEstimator` and :class:`sklearn.base.TransformerMixin`
    that means you can use it in a scikit-learn way.
    ::

        feature = SomeFeature()
        features = feature.fit_transform(X)

    .. note::

        The ``convert`` method should be rewrite to deal with single case. And the ``transform`` and ``fit_transform``
        Will be established for list of case automatically.

    **Adding references**

    ``BaseFeature`` also provide you to retrieving proper references for a feature.
    The ``__citations__`` returns a list of papers that should be cited.
    The ``__authors__`` returns a list of people who wrote the feature.
    Also can be accessed from property ``citations`` and ``citations``.

    These operations must be implemented for each new feature:

    - ``feature_labels`` - Generates a human-meaningful x_name for each of the features. Implement this as property.

    which can be set by ``set_feature_labels``

    Also suggest to implement these two properties:

    - ``citations`` - Returns a list of citations in BibTeX format.
    - ``authors`` - Returns a list of people who contributed writing a paper.

    .. note::

        None of these operations should change the state of the feature. I.e.,
        running each method twice should no produce different results, no class
        attributes should be changed, Running one operation should not affect the
        output of another.

    """

    __authors__ = ['boliqq07']
    __citations__ = ['No citations']
    _n_jobs = 1
    _feature_labels = []

    def __init__(self, n_jobs: int = 1, *, on_errors: str = 'raise', return_type: str = 'any',
                 batch_calculate: bool = False,
                 batch_size: int = 30):
        """
        Parameters
        ----------
        batch_size: int
            size of batch.
        batch_calculate :bool
            batch_calculate or not.
        n_jobs: int
            Parallel number.
        on_errors: str
            How to handle the exceptions in a feature calculations. Can be ``nan``, ``keep``, ``raise``.
            When 'nan', return a column with np.nan. The length of column corresponding to the number of feature
            labs.
            The default is 'raise' which will raise up the exception.
        return_type: str
            Specific the return type.
            Can be ``any``, ``array`` and ``df``.
            'array' and 'df' force return type to np.ndarray and pd.DataFrame respectively.
            If 'any', without type conversion .
            Default is 'any'
        """
        self.return_type = return_type
        self.n_jobs = n_jobs
        self.on_errors = on_errors
        self._kwargs = {}
        self.support_ = []
        self.d2 = False
        self._feature_labels = []
        self.batch_calculate = batch_calculate
        self.batch_size = batch_size
        import inspect
        inspect.getfullargspec(self.convert)

    @property
    def n_jobs(self):
        """
        n_jobs: int
            Parallel number.
        """
        return self._n_jobs

    @n_jobs.setter
    def n_jobs(self, n_jobs: int):
        """Set the number of threads for this"""
        if n_jobs > cpu_count() or n_jobs == -1:
            self._n_jobs = cpu_count()
        else:
            self._n_jobs = n_jobs

    def fit(self, *args, **kwargs):
        """fit function in :class:`BaseFeature` are weakened and just pass parameter."""
        _ = args
        self._kwargs = kwargs
        return self

    def fit_transform(self, X: List, y=None, **kwargs) -> Any:
        """
        If `convert` takes multiple inputs, supply inputs as a list of tuples.

        Copy from Mixin class for all transformers in scikit-learn. TransformerMixin

        Fit to data, then transform it.

        Fits transformer to `X` and `y` with optional parameters `fit_params`
        and returns a transformed version of `X`.

        Parameters
        ----------
        X : list
            list of case.
        y : None
            deprecated.
        **kwargs : dict
            Additional fit or transform parameters.

        Returns
        --------
        X_new :
            result data.
        """
        # non-optimized default implementation; override when a better
        # method is possible for a given clustering algorithm
        if y is None:
            # fit method of arity 1 (unsupervised transformation)
            return self.fit(X, **kwargs).transform(X)
        else:
            # fit method of arity 2 (supervised transformation)
            return self.fit(X, y, **kwargs).transform(X)

    def transform(self, entries: List) -> Any:
        """
        transform a list of entries.
        If `convert` takes multiple inputs, supply inputs as a list of tuples.

        Parameters
        ----------
        entries: list
            A list of entries to be featured.

        Returns
        ---------
        DataFrame:
            features for each entry.
        """

        # Check inputs
        if not isinstance(entries, Iterable):
            raise TypeError('parameter "entries" must be a iterable object')

        # Special case: Empty list
        if len(entries) == 0:
            return []

        # Run the actual feature
        rets = parallelize(self.n_jobs, self._wrapper, entries, tq=True)

        ret, self.support_ = zip(*rets)

        try:
            labels_len = len(self.feature_labels)
            if labels_len > 0:
                labels = self.feature_labels
            else:
                labels = None
        except (NotImplementedError, TypeError):
            labels = None

        if self.return_type == 'any':
            return ret

        if self.return_type == 'array':
            return np.array(ret)

        if self.return_type == 'df':
            if isinstance(entries, (pd.Series, pd.DataFrame)):
                return pd.DataFrame(ret, index=entries.index, columns=labels)
            return pd.DataFrame(ret, columns=labels)

    def _wrapper(self, x):
        """
        An exception wrapper for feature, used in feature_many and
        feature_dataframe. feature_wrapper changes the behavior of feature
        when ignore_errors is True in feature_many/dataframe.

        Parameters
        ----------
        x:
            input0 data_cluster to feature (type depends on feature).

        Returns
        -------
        result:tuple
            The first is calculated result, and the second is bool to mark data availability.
        """

        try:
            if isinstance(x, tuple):
                try:
                    con = self.convert(*x)
                except TypeError as e:
                    print(e)
                    raise TypeError("The tuple parameter would be separate automatically,"
                                    "if you case data is tuple, and Pass it to the first argument,"
                                    "please change it to list")
            else:
                try:
                    con = self.convert(x)
                except TypeError as e:
                    print(e)
                    raise TypeError("If the parameters should be unpack, please change the case to tuple, "
                                    "the other type(like list) would just pass to the first parameter integrally")

            if isinstance(con, (List, Tuple)):
                if len(con) == 2 and isinstance(con[1], bool):
                    pass
                else:
                    con = (con, True)
            else:
                con = (con, True)
            return con

        except BaseException as e:
            if self.on_errors == "nan":
                print("Bad conversion for:", x)
                return np.nan, False
            elif self.on_errors == "raise":
                raise e

    def convert(self, d):
        """
        Main feature function, which has to be implemented
        in any derived feature subclass.

        Parameters
        -----------
        d:
            input data_cluster to feature.

        Returns
        -------
        new_x: np.ndarray
            new x.
        conversion:bool
            Whether the conversion is perfect.
        """

        if isinstance(d, np.ndarray):
            if d.ndim == 1:
                return np.array(self._convert(d))
            if d.ndim == 2:
                if self.d2:
                    return np.array(self._convert(d))
                else:
                    return np.array([self._convert(i) for i in d])
            if d.ndim == 3:
                return np.array([[self._convert(i) for i in di] for di in d])

        elif isinstance(d, (list, tuple, int, float, dict)):
            return np.array(self._convert(d))
        raise NotImplementedError("just accept 1d,2d np.array or list, dict")

    @property
    def feature_labels(self):
        """Generate attribute names.

        Returns:
            ([str]) attribute labels.
        """
        return self._feature_labels

    def set_feature_labels(self, values: List[str]):
        """Generate attribute names.

        Returns:
            ([str]) attribute labels.
        """
        self._feature_labels = values

    @property
    def citations(self):
        """Citation(s) and reference(s) for this feature.

        Returns:
            (list) each element should be a string citation,
                ideally in BibTeX format.
        """
        return '\n'.join(self.__citations__)

    @property
    def authors(self):
        """List of implementors of the feature.

        Returns:
            (list) each element should either be a string with author x_name (e.g.,
                "Anubhav Jain") or a dictionary  with required key "x_name" and other
                keys like "email" or "institution" (e.g., {"x_name": "Anubhav
                Jain", "email": "ajain@lbl.gov", "institution": "LBNL"}).
        """

        return '\n'.join(self.__authors__)

    def _convert(self, d: np.ndarray) -> np.ndarray:
        """1d array input"""
        return d

    def __add__(self, other):
        raise TypeError("This method has no add")


Converter = BaseFeature  # old name


class DummyConverter(Converter):
    """
    Dummy converter as a placeholder, Do nothing.
    """

    def convert(self, d) -> np.ndarray:
        """
        Dummy convert, does nothing to input.

        Args:
            d (Any): input object

        Returns: d

        """
        d = np.array(d)

        return d.reshape(-1, 1) if d.ndim == 1 else d

    def _convert(self, d: np.ndarray) -> np.ndarray:
        pass


class ConverterCat(BaseFeature):
    """Pack the converters in to one unified approach.
    The same type Converter would merged and different would order to run.
    Thus, keeping the same type is next to each other!

    Examples
    ----------
    >>> tmps = ConverterCat(
    ...    AtomEmbeddingMap(),
    ...    AtomEmbeddingMap("ie.json")
    ...    AtomTableMap(search_tp="name"))
    >>> tmp.convert(x)

    """

    def __init__(self, *args: Converter, n_jobs: int = 1, on_errors: str = 'raise', return_type: str = 'any'):
        """

        Parameters
        ----------
        args: Converter
            List of Converter
        """
        super().__init__(n_jobs=n_jobs, on_errors=on_errors, return_type=return_type)
        self.args = self.sums(list(args))

    @staticmethod
    def sums(args):
        """SUM"""
        i = 0
        while i < len(args) - 1:
            try:
                args[i] = args[i] + args[i + 1]
                del args[i + 1]
            except (AssertionError, TypeError, IndexError, NotImplementedError):
                i += 1
        return args

    def convert(self, d):
        """convert batched"""
        data = []
        for ci in self.args:
            data.append(ci.convert(d))
        else:
            try:
                return np.concatenate(data, axis=-1)
            except ValueError:
                return data



import unittest

from monty.json import MSONable
from pymatgen.analysis.local_env import VoronoiNN

from featurebox.featurizers.envir.local_env import *
from featurebox.utils.look_json import get_marked_class


class AD(MSONable):
    def __init__(self, a, b):
        self.a = a
        self.b = b


class TestGraph(unittest.TestCase):
    def test_ser(self):
        ad = AD(a=5, b=VoronoiNN_D())
        #
        # ad = serialize(VoronoiNN(cutoff=5.0))
        afd = ad.as_dict()
        afj = ad.to_json()
        ad2 = ad.from_dict(afd)

    def test_ser2(self):
        a = get_marked_class(VoronoiNN, NNDict)
        print(a.__class__)


if __name__ == '__main__':
    unittest.main()



# import imp
# import os
# import platform
# import sys
# from importlib import util
# from pathlib import Path
#
# import importlib
# import torch
# import torch.utils.cpp_extension
#
#
# def import_module_from_library(module_name, path, is_python_module):
#     file, path, description = imp.find_module(module_name, [path])
#     # Close the .so file after load.
#     with file:
#         if is_python_module:
#             return imp.load_module(module_name, file, path, description)
#         else:
#             torch.ops.load_library(path)
#
# source = """
# #include <iostream>
# #include <torch/extension.h>
#
#
# using namespace torch::indexing;
# using namespace std;
#
# inline torch::Tensor merge_idxi_mean(const torch::Tensor &x, const torch::Tensor &idxi) {
#
#     return at::mean(x.index({idxi}), 0, true);
# }
#
# inline torch::Tensor merge_idxi_sum(const torch::Tensor &x, const torch::Tensor &idxi) {
#
#     return at::sum(x.index({idxi}), 0, true);
#
# }
#
# inline torch::Tensor merge_idxi_max(const torch::Tensor &x, const torch::Tensor &idxi) {
#
#     return std::get<0>(at::max(x.index({idxi}), 0, true));
#
# }
#
# inline torch::Tensor merge_idxi_min(const torch::Tensor &x, const torch::Tensor &idxi) {
#
#     return std::get<0>(at::min(x.index({idxi}), 0, true));
#
# }
#
#
# torch::Tensor merge_idx_mean(const torch::Tensor &x, const std::vector<torch::Tensor> &idx) {
#     std::vector<torch::Tensor> rx;
#     for (const torch::Tensor &item : idx) {
#         torch::Tensor rxi = at::mean(x.index({item}), 0, true);
#         rx.push_back(rxi);
#     }
#     return at::cat(rx, 0);}
#
# torch::Tensor merge_idx_sum(const torch::Tensor &x, const std::vector<torch::Tensor> &idx) {
#     std::vector<torch::Tensor> rx;
#     for (const torch::Tensor &item : idx) {
#         torch::Tensor rxi = at::sum(x.index({item}), 0, true);
#         rx.push_back(rxi);
#     }
#     return at::cat(rx, 0);}
#
# torch::Tensor merge_idx_max(const torch::Tensor &x, const std::vector<torch::Tensor> &idx) {
#     std::vector<torch::Tensor> rx;
#     for (const torch::Tensor &item : idx) {
#         torch::Tensor rxi = std::get<0>(at::max(x.index({item}), 0, true));
#         rx.push_back(rxi);
#     }
#     return at::cat(rx, 0);}
#
# torch::Tensor merge_idx_min(const torch::Tensor &x, const std::vector<torch::Tensor> &idx) {
#     std::vector<torch::Tensor> rx;
#     for (const torch::Tensor &item : idx) {
#         torch::Tensor rxi = std::get<0>(at::min(x.index({item}), 0, true));
#         rx.push_back(rxi);
#     }
#     return at::cat(rx, 0);}
#
# torch::Tensor merge_idx(const torch::Tensor &x, const std::vector<torch::Tensor> &idx, const string& merge_type="mean") {
#     std::vector<torch::Tensor> rx;
#     if (merge_type == "mean"){
#     for (const torch::Tensor &item : idx) {
#         torch::Tensor rxi = at::mean(x.index({item}), 0, true);
#         rx.push_back(rxi);
#     }}
#     else if (merge_type == "sum"){
#         for (const torch::Tensor &item : idx) {
#             torch::Tensor rxi = at::sum(x.index({item}), 0, true);
#             rx.push_back(rxi);
#         }}
#
#     else if (merge_type == "max"){
#         for (const torch::Tensor &item : idx) {
#             torch::Tensor rxi = std::get<0>(at::max(x.index({item}), 0, true));
#             rx.push_back(rxi);
#         }}
#     else {
#         for (const torch::Tensor &item : idx) {
#             torch::Tensor rxi = std::get<0>(at::min(x.index({item}), 0, true));
#             rx.push_back(rxi);
#         }}
#
# //    auto rxx = torch::TensorList(rx);
#     return at::cat(rx, 0);
# }
#
# """
#
# name = "segment_method"
#
# if platform.system() == "Windows":
#     ext = "dll"
# else:
#     ext = "so"
#
# name_dir = "cache_" + name
# MODULE_DIR = Path(__file__).parent.absolute()
# MODULE_DIR_NAME_DIR = MODULE_DIR / name_dir
#
# if os.path.isdir(MODULE_DIR_NAME_DIR) and os.path.isfile(MODULE_DIR_NAME_DIR / "{}.{}".format(name, ext))\
#         :
#     mod = import_module_from_library(name, MODULE_DIR_NAME_DIR, True)
#
# else:
#     if not os.path.isdir(MODULE_DIR_NAME_DIR):
#         os.mkdir(MODULE_DIR_NAME_DIR)
#     mod = torch.utils.cpp_extension.load_inline(
#         name=name,
#         cpp_sources=source,
#         verbose=True,
#         build_directory=MODULE_DIR_NAME_DIR,
#         is_python_module=True,
#         functions=["merge_idx", "merge_idx_mean", "merge_idx_max", "merge_idx_sum", "merge_idx_min"]
#     )
#
# # if __name__ == '__main__':
# #     inputs = torch.tensor([1.2, 3, 4, 5, 6, 0.7, 9], requires_grad=True, device="cpu")
# #     node_bond_idx = [torch.tensor([1, 2, 3]), torch.tensor([4, 5, 6])]
#     # b = mod.merge_idx(inputs, node_bond_idx, "mean")



from abc import abstractmethod
from itertools import zip_longest
from pathlib import Path
from typing import List, Sequence, Iterable, Tuple

import numpy as np
import torch
from torch import Tensor
from torch.utils.data import Dataset, DataLoader
from torch.utils.data._utils.collate import default_convert

MODULE_DIR = Path(__file__).parent.absolute()


class _BaseGraphSingleGenerator(Dataset):

    def __init__(
            self,
            dataset_size: int,
            targets: np.ndarray,
            sample_weights: np.ndarray = None,
    ):
        """
        Args:
            dataset_size (int): Number of entries in dataset
            targets (ndarray): Feature to be predicted for each network
            sample_weights (npdarray): sample weights
        """
        if targets is not None:
            self.targets = np.array(targets).reshape((dataset_size, -1))
        else:
            self.targets = None

        if sample_weights is not None:
            self.sample_weights = np.array(sample_weights)
        else:
            self.sample_weights = None

        self.total_n = dataset_size
        self.mol_idx = np.arange(self.total_n)

    def __len__(self) -> int:
        return self.total_n

    def __iter__(self):
        """Create a generator that iterate over the Sequence."""
        for item in (self[i] for i in range(len(self))):
            yield item

    @staticmethod
    def _reform_data(
            *args: np.ndarray
    ) -> List:
        """
        1.add length of node_atom_idx and len of node_ele_idx,ele_atom_idx.
        2.transform the preprocessing to numpy with Fixed accuracy float32 and int64 .
        """

        def num(ax):
            ls = []
            temp = ax[0]
            li = -1
            for ai in range(ax.shape[0]):
                if np.all(np.equal(temp, ax[ai])):
                    li += 1
                else:
                    temp = ax[ai]
                    li += 1
                    ls.append(li)
                    li = 0
            ls.append(li + 1)
            return ls

        inputs = []

        for i in args:
            try:
                inputs.append(np.array(i, dtype=np.float32))
            except ValueError:
                inputs.append(i)

        inputs.append(args[0].shape[0])
        inputs.append(np.array(num(args[0])))
        inputs.append(inputs[-1].shape[0])
        return inputs

    def __getitem__(self, index: int) -> tuple:
        """please make sure always return a tuple!"""
        # Get the indices for this batch

        # Get the inputs for each batch
        inputs = self._generate_inputs(index)

        # Make the graph preprocessing
        inputs = self._reform_data(*inputs)

        # Return the batch
        if self.targets is None:
            return inputs,
        # get targets
        target_temp = self.targets[index]
        if self.sample_weights is None:
            return inputs, target_temp
        sample_weights_temp = self.sample_weights[index]

        return inputs, target_temp, sample_weights_temp

    def __setitem__(self, key, value):
        raise NotImplemented("This method is deleted in Dataset class.")

    @abstractmethod
    def _generate_inputs(self, batch_idx: int) -> Iterable:
        """Get the preprocessing by index"""


class GraphGenerator(_BaseGraphSingleGenerator):
    """
    A generator class that assembles several structures (indicated by
    batch_size) and form (x, y) pairs for model training.

    return
    [
    "atom_fea",
    "nbr_fea",
    "state_fea",
    "atom_nbr_idx",
    ...
    "node_atom_idx",
    "node_ele_idx"
    "ele_atom_idx"]
    """

    def __init__(
            self,
            atom_fea: List,
            nbr_fea: List[np.ndarray],
            state_fea: List[np.ndarray],
            atom_nbr_idx: List,
            *args,
            targets: np.ndarray = None,
            sample_weights: np.ndarray = None,
            **kwargs,
    ):
        """
        Base class has five probs, and not limit the number of attributes.

        More props would pass to arg, and named "other_prop_i_feature"
        or props would pass to kwarg, and named "other_prop_{...}_feature"

        Args:
            atom_fea: (list of np.array) list of atom feature matrix,
            nbr_fea: (list of np.array) list of bond features matrix
            state_fea: (list of np.array) list of [1, G] state features,
                where G is the global state feature dimension
            M is different for different structures
            atom_nbr_idx: (list of integer) list of (M, ) the other side atomic
                index of the bond, M is different for different structures,
                but it has to be the same as the corresponding index1.
            targets: (numpy array), N*1, where N is the number of structures
            sample_weights: (numpy array), N*1, where N is the number of structures
        """
        super().__init__(
            len(atom_fea), targets, sample_weights=sample_weights,
        )

        self.atom_fea = atom_fea
        self.nbr_fea = nbr_fea
        self.state_fea = state_fea

        self.atom_nbr_idx = atom_nbr_idx

        self.final_data_name = ["atom_fea", "nbr_fea", "state_fea", "atom_nbr_idx",
                                "node_atom_idx", "node_ele_idx", "ele_atom_idx"]
        self.add_prop = []
        for i, j in enumerate(args):
            name = "other_prop_{}_feature".format(i)
            self.__setattr__(name, j)
            self.final_data_name.append(name)
            self.add_prop.append(name)

        for i, j in kwargs.items():
            name = "other_prop_{}_feature".format(i)
            self.__setattr__(name, j)
            self.final_data_name.append(name)
            self.add_prop.append(name)

    def _generate_inputs(self, index: int):
        # Get the features and connectivity lists for this batch
        feature_list_temp = self.atom_fea[index]
        connection_list_temp = self.nbr_fea[index]
        global_list_temp = self.state_fea[index]
        atom_nbr_idx = self.atom_nbr_idx[index]

        return_list = [feature_list_temp, connection_list_temp, global_list_temp, atom_nbr_idx]

        for i in self.add_prop:
            try:
                return_list.append(self.__getattribute__(i)[index])
            except BaseException:
                raise TypeError("The {} is can not iterable".format(i))
        return return_list


def add_mark(data):
    """Change length to rank"""
    if isinstance(data, Tensor):
        data = data.numpy()
    if isinstance(data, list):
        data = np.array(data)
    a = np.cumsum(data)
    b = np.insert(a, 0, 0)[:-1]
    return [torch.arange(k, v) for k, v in zip(b, a)]


# def get_bond_node_idx(node_ele_idx, index1: Tensor):
#     spil = [index1[i] for i in node_ele_idx]
#     res = []
#     st = 0
#     for si in spil:
#         res.extend([torch.where(si == i)[0] + st for i in range(int(torch.max(si)) + 1)])
#         st += len(si)
#     res = [i.to(torch.int64) for i in res]
#     return res


class MGEDataLoader:
    """
    This loader:
        1.add the ability to collate preprocessing.
        2.Process and add the necessary indexes list.

    Notes:
        Though, the collate_fn parameter in torch.DataLoader could be used to collate preprocessing,
        but when we try to use it, we find it make the code difficult to read and maintain.
        thus, we just use "collate_fn=default_convert" to transform the preprocessing from numpy to tensor,
        and collate preprocessing outer of the torch.DataLoader

    Notes:
        we replace the last 2 features in primary features to index list feature and add 1 features.

        len_node_atom -> node_atom_idx

        len_node_ele -> node_ele_idx

        ele_atom_idx -> node_ele_idx

    Thus the number of final features are add 1 to input dataset.

    CrystalGraphDisordered,CrystalGraph,SingleMoleculeGraph in crystal, return 7 features.
    Return:
    ["atom_feature", "nbr_feature", "state_feature", 'atom_nbr_idx',
    "node_atom_idx", "node_ele_idx","ele_atom_idx"]

    other Graph self-defined, Return:
    ["atom_feature", "nbr_feature", "state_feature", 'atom_nbr_idx',
    ...
    "node_atom_idx", "node_ele_idx","ele_atom_idx"]

    """

    def __init__(self, dataset: GraphGenerator, collate_marks=None, batch_size=10, shuffle=True, num_workers=0,
                 **kwargs):
        """
        collate_marks:
                Transform all np.array to tensor, by stack or cat with self-defined,
                due to the elements of preprocessing in batch dont have consistent size.\n

                **Default is ('c', 'c', 's', 'c', 'f', 'c', 'f')**

                atom_fea "c",
                nbr_fea "c",
                state_fea "s",
                atom_nbr_idx = "c",

                ... other feature your defined.

                node_atom_idx = "f",
                node_ele_idx = "c",
                ele_atom_idx = "f",

        Args:
            dataset:(GraphGenerator)
            batch_size:(int)
            shuffle:(bool)
            num_workers:(int)
            kwargs:(Any) parameter for torch.DataLoader
            collate_marks:(tuple,list)

        """

        self.loader = DataLoader(
            dataset=dataset, batch_size=batch_size, shuffle=shuffle, num_workers=num_workers,
            collate_fn=default_convert, **kwargs)
        if collate_marks is None:
            collate_marks = ('c', 'c', 's', 'c', 'f', 'c', 'f')
        self.collate_marks = tuple(collate_marks)
        self._number_yield = 0
        self.device = torch.device('cpu')

    def __iter__(self):
        return self

    def __len__(self):
        return self.loader.__len__()

    def to_cuda(self, device='cuda:0'):
        """Change the data to cuda"""
        self.device = torch.device(device)

    def move(self, batch):
        """move data to cuda"""
        ba = []
        for i in batch:
            if isinstance(i, Tensor):
                ba.append(i.to(self.device))
            elif isinstance(i, Sequence):
                ba.append(self.move(i))
            else:
                raise TypeError("just accept list tuple of tensors or tensor")
        return ba

    def __next__(self):
        if self._number_yield < self.loader.__len__():
            batch = next(self.loader.__iter__())
            batch = self.collate_fn(batch, self.collate_marks)
            batch[0] = self.index_transform(batch[0])
            batch = self.move(batch)
            self._number_yield += 1
            return batch
        raise StopIteration

    def reset(self):
        self._number_yield = 0
        self.loader._iterator = self.loader._get_iterator()

    @staticmethod
    def collate_fn(batch: List, collate_marks: Tuple[str]) -> List[List]:
        """
        User collate function.

        Args:
            collate_marks: (tuple of str)
                the same size with features
            batch: (list)
                shape N_sample.
                Each one is a list or tuple,
                such as [[f1i,f2i],] or [[f1i,f2i], yi_value ] or [[f1i,f2i],yi_value, weight_i]

        Returns:
            such as [[f1,f2],] or [[f1,f2],y] or [[f1,f2],y,w]
        """

        def _deal_func(a, di="c"):
            """

            Args:
                a: (float,int,Tensor)
                    one feature with list format.
                di: str
                    collate type

            Returns:
                Tensor or list
            """
            elem = a[0]
            if isinstance(elem, float):
                a = torch.tensor(a, dtype=torch.float32)
                return torch.reshape(a, (-1, 1))
            elif isinstance(elem, int):
                a = torch.tensor(a, dtype=torch.int64)
                return torch.reshape(a, (-1, 1))
            elif isinstance(elem, Tensor):
                if elem.shape == ():
                    a = torch.stack(a, 0)
                    return torch.reshape(a, (-1, 1)) if len(a.shape) == 1 else a
                if di in ["c", "cat"]:
                    a = torch.cat(a, 0)
                    return torch.reshape(a, (-1, 1)) if len(a.shape) == 1 else a
                elif di in ["s", "stack"]:
                    a = torch.stack(a, 0)
                    return torch.reshape(a, (-1, 1)) if len(a.shape) == 1 else a
                else:
                    return a
            else:
                raise TypeError("just accept: Tensor, int, float")

        assert len(batch[0][0]) == len(
            collate_marks) or len(batch[0][0]) == len(
            collate_marks) + 3, "Must give one 'collate_marks' parameter and the shape is same with features.\n" \
                                "such as feature = [f1,f2,f3,f4,f5,f6,f7,f8,f9], without target" \
                                "The collate_marks = ('c', 'c', 's', 'c', 'f', 'c','f')\n" \
                                "'s' is torch.stack, 'c' is torch.cat, 'f' do noting but just return the list"
        transposed = list(zip(*batch))
        return [[_deal_func(si, di) for si, di in zip_longest(zip(*samples), collate_marks, fillvalue="f")]
                if i == 0 else _deal_func(samples, "c") for i, samples in enumerate(transposed)]

    @staticmethod
    def index_transform(data):
        """
        Transform the length of node_atom_idx and node_ele_idx to list of index.
        and add new node_bond_idx.

        [1,3,7,...] to  [[0,], [1,2,3], [4,5,6,7,8,9,10],...]

        Args:
            data: (List)
                batch preprocessing without collate.

        Returns:
            batch preprocessing with collate

        """

        data[-3] = add_mark(data[-3])  # node_atom_idx
        data[-2] = add_mark(data[-2])  # node_ele_idx
        data[-1] = add_mark(data[-1])  # ele_atom_idx

        data[3] = data[3].to(torch.int64)  # atom_nbr_idx
        if len(data[1].shape) == 2:
            data[1] = data[1].unsqueeze(2)  # nbr_fea

        return data



#!/usr/bin/python3.7
# -*- coding: utf-8 -*-

# @Time   : 2019/7/27 16:57
# @Author : Administrator
# @Software: PyCharm
# @License: BSD 3-Clause

"""
Calculate the the correction of columns.
"""
import copy
from typing import List

import numpy as np
from mgetool.tool import name_to_name
from sklearn.base import BaseEstimator, MetaEstimatorMixin
from sklearn.datasets import load_boston
from sklearn.feature_selection import SelectorMixin
from sklearn.utils import check_random_state
from sklearn.utils.validation import check_is_fitted

from featurebox.selection.mutibase import MutiBase


class Corr(BaseEstimator, MetaEstimatorMixin, SelectorMixin, MutiBase):
    """
    Calculate correlation. (Where the result are changed with random state.)

    Examples
    ---------
    >>> from sklearn.datasets import load_boston
    >>> from featurebox.selection.corr import Corr
    >>> x, y = load_boston(return_X_y=True)
    >>> co = Corr(threshold=0.5)
    >>> nx = co.fit_transform(x)

    **1. Used for get group exceeding the threshold**

    Examples
    ---------

    >>> from sklearn.datasets import load_boston
    >>> from featurebox.selection.corr import Corr
    >>> x, y = load_boston(return_X_y=True)
    >>> co = Corr(threshold=0.7)
    >>> groups = co.count_cof(np.corrcoef(x[:,:7], rowvar=False))
    >>> groups[1]
    [[0], [1], [2, 4], [3], [2, 4, 6], [5], [4, 6]]
    >>> groups[0]
    [[1.0], [1.0], [1.0, 0.764], [1.0], [0.764, 1.0, 0.731], [1.0], [0.731, 1.0]]

    Where the (2,4), (2,4,6), (4,6) are with correlation more than 0.7.

    **2. Used for filter automatically by machine**

    Examples
    -----------
    >>> from sklearn.datasets import load_boston
    >>> from featurebox.selection.corr import Corr
    >>> x, y = load_boston(return_X_y=True)
    >>> co = Corr(threshold=0.7)
    >>> co.fit(x)
    Corr(threshold=0.7)
    >>> group = co.count_cof()
    >>> group[1]
    [[0], [1], [2, 4, 7, 9], [3], [2, 4, 6, 7], [5], [4, 6, 7], [2, 4, 6, 7], [8, 9], [2, 8, 9], [10], [11], [12]]
    >>> co.remove_coef(group[1]) # Filter automatically by machine.
    [0, 1, 2, 3, 5, 6, 8, 10, 11, 12]

    Where the remove_coef are changed with random state.

    **3. Used for binding correlation**

    Examples
    -----------
    >>> from sklearn.datasets import load_boston
    >>> from featurebox.selection.corr import Corr
    >>> x, y = load_boston(return_X_y=True)
    >>> co = Corr(threshold=0.7,muti_index=[0,8],muti_grade=2)
    >>> # in range [0,8], the features are binding in to 2 sized: [[0,1],[2,3],[4,5],[6,7]]
    >>> co.fit(x)
    Corr(muti_index=[0, 8], threshold=0.7)
    >>> group = co.count_cof()
    >>> group[1]
    [[0], [1], [2], [3], [4, 5], [4, 5], [6], [7], [8]]
    >>> co.remove_coef(group[1]) # Filter automatically by machine.
    [0, 1, 2, 3, 4, 6, 7, 8]

    Where 4 is filtered , Corresponding to the initial feature 8.

    [0,1] -> 0; [2,3] -> 1; [4,5]->2; [6,7]->3, 8->4; 9->5; 10->6; 11->7; 12->8; 13->9;
    """

    def __init__(self, threshold: float = 0.85, muti_grade: int = 2, muti_index: List = None, must_index: List = None,
                 random_state: int = 0):
        """

        Parameters
        ----------
        threshold:float
            ranking threshold.
        muti_grade:
            binding_group size, calculate the correction between binding.
        muti_index:list
            the range of muti_grade:[min,max).
        must_index:list
            the columns force to index.
        random_state:int
            int
        """

        super().__init__(muti_grade=muti_grade, muti_index=muti_index, must_index=must_index)
        self.threshold = threshold
        self.cov = None
        self.cov_shrink = None
        self.shrink_list = []
        self.random_state = random_state

    def fit(self, data, pre_cal=None, method="mean"):
        if pre_cal is None:

            cov = np.corrcoef(data, rowvar=False, )

        elif isinstance(pre_cal, np.ndarray) and pre_cal.shape[0] == data.shape[1]:
            cov = pre_cal
        else:
            raise TypeError("pre_cal is None or coef of data_cluster with shape(data_cluster[0],data_cluster[0])")
        cov = np.nan_to_num(cov - 1) + 1
        self.cov = cov
        self.data = data
        self.shrink_list = list(range(self.cov.shape[0]))
        self._shrink_coef(method=method)
        self.filter()
        return self

    def _shrink_coef(self, method="mean" or "max"):

        if self.check_muti:
            self.shrink_list = list(range(self.cov.shape[0]))
            self.shrink_list = list(self.feature_fold(self.shrink_list))

            cov = self.cov
            single = tuple([i for i in self.shrink_list if i not in self.check_muti])
            muti = tuple([i for i in self.shrink_list if i in self.check_muti])

            cov_muti_all = []
            le = self.muti_grade
            while le:
                index = []
                index.extend(single)
                index.extend([i + le - 1 for i in muti])
                index.sort()
                cov_muti_all.append(cov[index][:, index])
                le -= 1
            cov_muti_all = np.array(cov_muti_all)
            if method == "mean":
                cov_new = np.mean(cov_muti_all, axis=0)
            else:
                cov_new = np.max(cov_muti_all, axis=0)
            self.cov_shrink = cov_new
            return self.cov_shrink
        else:
            self.cov_shrink = self.cov

    def count_cof(self, cof=None):
        """Check cof and count the number."""
        if cof is None:
            cof = self.cov_shrink
        if cof is None:
            raise NotImplemented("imported cof is None")

        list_count = []
        list_coef = []
        g = np.where(abs(cof) >= self.threshold)
        for i in range(cof.shape[0]):
            e = np.where(g[0] == i)
            com = list(g[1][e])
            # ele_ratio.remove(i)
            list_count.append(com)
            list_coef.append([round(cof[i, j], 3) for j in com])
        self.list_coef = list_coef
        self.list_count = list_count
        return list_coef, list_count

    def remove_coef(self, cof_list_all):
        """Delete the index of feature with repeat coef."""
        ran = check_random_state(self.random_state)
        reserve = []
        for i in cof_list_all:
            if not cof_list_all:
                reserve.append(i)

        for cof_list in cof_list_all:
            if not cof_list:
                pass
            else:
                if reserve:
                    candi = []
                    for j in cof_list:

                        con = any([[False, True][j in cof_list_all[k]] for k in reserve])
                        if not con:
                            candi.append(j)
                    if any(candi):
                        a = ran.choice(candi)
                        reserve.append(a)
                    else:
                        pass
                else:
                    a = ran.choice(cof_list)
                    reserve.append(a)
                cof_list_t = copy.deepcopy(cof_list)
                for dela in cof_list_t:
                    for cof_list2 in cof_list_all:
                        if dela in cof_list2:
                            cof_list2.remove(dela)
        return sorted(list(set(reserve)))

    def filter(self):
        list_coef, list_count = self.count_cof()
        index = self.remove_coef(list_count)
        support_ = [self.shrink_list[i] for i in index]
        support_ = self.feature_unfold(support_)
        self.support_ = np.array([True if i in support_ else False for i in range(self.data.shape[1])])
        return support_

    @staticmethod
    def cov_y(x_, y_):
        cov = np.corrcoef(x_, y_, rowvar=False)
        cov = cov[:, -1][:-1]
        return cov

    def remove_by_y(self, y_):
        corr = self.cov_y(self.data, y_)
        corr = self.feature_fold(corr)
        lcount = self.list_count
        fea_all = []
        score = name_to_name(corr, search=lcount, search_which=0, return_which=(1,), two_layer=True)
        for score_i, list_i in zip(score, lcount):
            indexs = np.argmax(score_i)
            feature_index = list_i[int(indexs)]
            fea_all.append(feature_index)

        fea_all = sorted(list(set(fea_all)))
        return fea_all

    def _get_support_mask(self):
        check_is_fitted(self, 'support_')
        return self.support_

    # def transform_index(self, data):
    #     if isinstance(data, int):
    #         return self.shrink_list.index(data)
    #     elif isinstance(data, (list, tuple)):
    #         return [self.shrink_list.index(i) for i in data]
    #
    # def inverse_transform_index(self, data):
    #     if isinstance(data, int):
    #         return self.shrink_list[data]
    #     elif isinstance(data, (list, tuple)):
    #         return [self.shrink_list[i] for i in data]
    #     else:
    #         pass
    #
    # def transform(self, data):
    #     if isinstance(data, (list, tuple)):
    #         return data[self.shrink_list]
    #     elif isinstance(data, np.ndarray) and data.ndim == 1:
    #         return data[self.shrink_list]
    #     elif isinstance(data, np.ndarray) and data.ndim == 2:
    #         return data[:, self.shrink_list]
    #     else:
    #         pass


if __name__ == "__main__":
    # x, y = load_boston(return_X_y=True)
    # co = Corr(threshold=0.7)
    # c = co.count_cof(np.corrcoef(x, rowvar=False))[1]

    x, y = load_boston(return_X_y=True)
    co = Corr(threshold=0.5, muti_index=[0, 8], muti_grade=2)

    nx = co.fit_transform(x)



import math

import numpy as np
from ase.neighborlist import NeighborList


class EAMD:
    """EAMD is an atom-centered descriptor that is inspired by Embedded Atom method (EAM).
    The EAM utilizes the orbital-dependent density components. The orbital-dependent
    component consists of a set of local atomic density descriptions.

    The functional form of EAMD is consistent with:
        Zhang, Y., et. al. (2019). The Journal of Physical Chemistry Letters, 10(17), 4962-4967.

    Parameters
    ----------
    parameters: dict
        The user-defined parameters for component of local atomic density descriptions.
        i.e. {'L': 2, 'eta': [0.36], 'Rs': [1.0]}
    Rc: float
        The EAMD will be calculated within this radius.
    derivative: bool
        If True, calculate the derivative of EAMD.
    stress: bool
        If True, calculate the virial stress contribution of EAMD.
    """

    def __init__(self, parameters=None, Rc=5., derivative=True, stress=False):
        if parameters is None:
            parameters = {'L': 2, 'eta': [0.36], 'Rs': [1.0]}
        self._type = 'EAMD'
        self.dsize = int(1)
        self.parameters = {}

        # Set up keywords
        keywords = ['L', 'eta', 'Rs']
        for k, v in parameters.items():
            if k not in keywords:
                msg = f"{k} is not a valid key. " \
                      f"Choose from {keywords}"
                raise NotImplementedError(msg)
            else:
                if k == 'L':
                    self.dsize *= (v + 1)
                    self.parameters[k] = v
                else:
                    self.dsize *= len(v)
                    self.parameters[k] = np.array(v)

        self.Rc = Rc
        self.derivative = derivative
        self.stress = stress

    def calculate(self, crystal, ids=None):
        """Calculate and return the EAMD.
        
        Parameters
        ----------
        crystal: object
            ASE Structure object.
        ids: list
            A list of the centered atoms to be computed
            if None, all atoms will be considered
        
        Returns
        -------
        d: dict
            The user-defined EAMD that represent the crystal.
            d = {'x': [N, d], 'dxdr': [N, m, d, 3], 'rdxdr': [N, m, d, 3, 3],
            'elements': list of elements}
        """
        self.crystal = crystal
        self.total_atoms = len(crystal)  # total atoms in the unit cell
        vol = crystal.get_volume()

        rc = [self.Rc / 2.] * self.total_atoms
        neighbors = NeighborList(rc, self_interaction=False, bothways=True, skin=0.)
        neighbors.update(crystal)

        unique_N = 0
        for i in range(self.total_atoms):
            indices, offsets = neighbors.get_neighbors(i)
            ith = 0
            if i not in indices:
                ith += 1
            unique_N += len(np.unique(indices)) + ith  # +1 is for i

        # Make numpy array here
        self.d = {'x': np.zeros([self.total_atoms, self.dsize]), 'elements': []}
        if self.derivative:
            self.d['dxdr'] = np.zeros([unique_N, self.dsize, 3])
            self.d['seq'] = np.zeros([unique_N, 2], dtype=int)
        if self.stress:
            self.d['rdxdr'] = np.zeros([unique_N, self.dsize, 3, 3])

        seq_count = 0

        if ids is None:
            ids = range(len(crystal))

        for i in ids:  # range(self.total_atoms):
            element = crystal.get_chemical_symbols()[i]
            indices, offsets = neighbors.get_neighbors(i)
            Z = []  # atomic numbers of neighbors

            assert len(indices) > 0, \
                f"There's no neighbor for this structure at Rc = {self.Rc} A."

            Ri = crystal.get_positions()[i]
            total_neighbors = len(indices)

            Rj = np.zeros([total_neighbors, 3])
            IDs = np.zeros(total_neighbors, dtype=int)

            count = 0
            for j, offset in zip(indices, offsets):
                Rj[count, :] = crystal.positions[j] + np.dot(offset, crystal.get_cell())
                IDs[count] = j
                Z.append(crystal[j].number)
                count += 1
            Z = np.array(Z)

            Rij = Rj - Ri
            Dij = np.sqrt(np.sum(Rij ** 2, axis=1))

            d = calculate_eamd(i, self.total_atoms, Rij, Dij, Z, IDs, self.Rc,
                               self.parameters, self.derivative, self.stress)

            self.d['x'][i] = d['x']
            if self.derivative:
                n_seq = len(d['seq'])
                self.d['dxdr'][seq_count:seq_count + n_seq] = d['dxdr']
                self.d['seq'][seq_count:seq_count + n_seq] = d['seq']
            if self.stress:
                self.d['rdxdr'][seq_count:seq_count + n_seq] = -d['rdxdr'] / vol

            if self.derivative:
                seq_count += n_seq

            self.d['elements'].append(element)

        return self.d


def calculate_eamd(i, m, rij, dij, Z, IDs, Rc, parameters, derivative, stress):
    """ Calculate the EAMD for a center atom i.
    
    Parameters
    ----------
    i: int
        The i-th atom center.
    m: int
        The total atoms in the crystal unit cell.
    rij: array [j, 3]
        The vector distances of atom i to neighbors j.
    dij: array [j]
        The array of distances of i-th center atom.
    Z: array [j]
        The atomic numbers of neighbors.
    IDs: int array [j]
        The indices of neighbors centering about atom i.
    Rc: float
        The cutoff radius.
    parameters: dict
        Rs: float array (d1)
            The shift from the center of the Gaussian-type orbitals.
        etas: float array (d2)
            The width of the Gaussian-type orbitals.
        L: int (d3)
            The total orbital angular momentum.
    derivative:
        If True, calculate the derivative of EAMD.
    stress: bool
        If True, calculate the virial stress contribution of EAMD.

    Returns
    -------
    Dict of EAMD descriptors with its derivative and stress contribution.
    """
    l_index = [1, 4, 10, 20]
    normalize = 1 / np.sqrt(np.array([1, 1, 1, 1, 1, 1, 1, 2, 2, 2,
                                      1, 2, 2, 2, 2, 2, 2, 6, 6, 6]))

    Rs = parameters['Rs']  # d1
    etas = parameters['eta']  # d2
    L = parameters['L']  # d3
    d1, d2, d3, j = len(Rs), len(etas), L + 1, len(dij)

    ij_list = i * np.ones([len(IDs), 2], dtype=int)
    ij_list[:, 1] = IDs

    unique_js = np.unique(IDs)
    if i not in unique_js:
        unique_js = np.append(i, unique_js)
    unique_js.sort()
    seq = i * np.ones([len(unique_js), 2], dtype=int)
    seq[:, 1] = unique_js
    uN = len(unique_js)
    _i = np.where(unique_js == i)[0][0]

    term1, d_term1 = get_xyz(unique_js, rij, ij_list, L, derivative=derivative)  # [j, D3], [j, uN, 3, l]

    d0 = dij - Rs[:, np.newaxis]
    d02 = d0 ** 2  # [d1, j]

    fc = Cosine(dij, Rc)  # [j]
    cj_cutoff = Z * fc  # [j]
    term2_1 = np.exp(np.einsum('i,jk->ijk', -etas, d02))  # [d2, d1, j]
    term2 = np.einsum('ijk,k->ijk', term2_1, cj_cutoff)  # [d2, d1, j]

    term = np.einsum('ij,kli->jkl', term1, term2)  # [D3, d2, d1]

    if derivative:
        dterm0 = np.einsum('k, ijk->ijk', Z, term2_1).reshape([d1 * d2, j])  # [d2*d1, j]
        dterm11 = np.einsum('ij, j->ij', dterm0, fc).reshape([d1 * d2, j])  # [d2*d1, j]
        dterm1 = np.einsum('ij, jklm->jmilk', dterm11, d_term1)  # [j, D3, d2*d1, 3, uN]

        dterm20 = np.einsum('ij, ki->jki', term1, dterm0)  # [D3, d2*d1, j]
        dterm21 = CosinePrime(dij, Rc)  # [j]
        _dterm22 = np.einsum('ij,j->ij', d0, fc)  # [d1, j]
        dterm22 = 2 * np.einsum('i,jk->ijk', etas, _dterm22)  # [d2, d1, j]
        dterm23 = (dterm21 - dterm22).reshape([d2 * d1, j])  # [d2*d1, j]
        dterm24 = np.einsum('ijk, jk->ijk', dterm20, dterm23)  # [D3, d2*d1, j]

        dRij_dRm = np.zeros([j, 3, uN])
        for mm, _m in enumerate(unique_js):
            mm_list = _m * np.ones([j, 1], dtype=int)
            dRij_dRm[:, :, mm] = dRij_dRm_norm(rij, np.hstack((ij_list, mm_list)))  # [j, 3, uN]

        dterm2 = np.einsum('ijk, klm->kijlm', dterm24, dRij_dRm)  # [j, D3, d2*d1, 3, uN]

        dphi_dRm = dterm1 + dterm2  # [j, D3, d2*d1, 3, uN]
        dterm = np.einsum('ij, hijkl->ijkl', term.reshape([term.shape[0], d2 * d1]), dphi_dRm)  # [D3, d2*d1, 3, uN]

        if stress:
            _RDXDR = np.zeros([term.shape[0], d2 * d1, 3, uN, 3])  # [D3, d2*d1, 3, uN, 3]
            for count, ij in enumerate(ij_list):
                # _j = ij[1]
                _j = np.where(unique_js == ij[1])[0][0]
                tmp = dphi_dRm[count, :, :, :, _j]
                _RDXDR[:, :, :, _j, :] += np.einsum('ijk,l->ijkl', tmp, rij[count])

            sterm = np.einsum('ij, ijklm->ijklm', term.reshape([term.shape[0], d2 * d1]), _RDXDR)

    count = 0
    x = np.zeros([d3 * d2 * d1])  # [d3*d2*d1]
    dxdr, rdxdr = None, None
    if derivative:
        dxdr = np.zeros([uN, d1 * d2 * d3, 3])  # [uN, d3*d2*d1, 3]
    if stress:
        rdxdr = np.zeros([uN, d1 * d2 * d3, 3, 3])
    for l in range(L + 1):
        Rc2l = Rc ** (2 * l)
        L_fac = math.factorial(l)
        x[count:count + d1 * d2] = L_fac * np.einsum('ijk->jk', term[:l_index[l]] ** 2).ravel() / Rc2l
        if derivative:
            dxdr[:, count:count + d1 * d2, :] = 2 * L_fac * np.einsum('ijkl->ljk', dterm[:l_index[l]]) / Rc2l
        if stress:
            rdxdr[:, count:count + d1 * d2, :, :] = 2 * L_fac * np.einsum('ijklm->ljkm', sterm[:l_index[l]]) / Rc2l

        count += d1 * d2

    return {'x': x, 'dxdr': dxdr, 'rdxdr': rdxdr, 'seq': seq}


def get_xyz(unique_js, rij, ij_list, L, derivative):
    """ (x ** l_x) * (y ** l_y) * (z ** l_z) / (l_x! * l_y! * l_z!) ** 0.5 """
    normalize = 1 / np.sqrt(np.array([1, 1, 1, 1, 1, 1, 1, 2, 2, 2,  # 1 / sqrt(lx! ly! lz!)
                                      1, 2, 2, 2, 2, 2, 2, 6, 6, 6]))

    L_list = [[[0], [0]],  # lx = 1, ly = 0, lz = 0; L = 1
              [[1], [1]],  # lx = 0, ly = 1, lz = 0; L = 1
              [[2], [2]],  # lx = 0, ly = 0, lz = 1; L = 1
              [[0, 1], [0, 1]],  # lx = 1, ly = 1, lz = 0; L = 2
              [[0, 2], [0, 2]],  # lx = 1, ly = 0, lz = 1; L = 2
              [[1, 2], [1, 2]],  # lx = 0, ly = 1, lz = 1; L = 2
              [[0], [3]],  # lx = 2, ly = 0, lz = 0; L = 2
              [[1], [4]],  # lx = 0, ly = 2, lz = 0; L = 2
              [[2], [5]],  # lx = 0, ly = 0, lz = 2; L = 2
              [[0, 1, 2], [0, 1, 2]],  # lx = 1, ly = 1, lz = 1; L = 3
              [[1, 2], [1, 5]],  # lx = 0, ly = 1, lz = 2; L = 3
              [[1, 2], [4, 2]],  # lx = 0, ly = 2, lz = 1; L = 3
              [[0, 2], [0, 5]],  # lx = 1, ly = 0, lz = 2; L = 3
              [[0, 1], [0, 4]],  # lx = 1, ly = 2, lz = 0; L = 3
              [[0, 1], [3, 1]],  # lx = 2, ly = 1, lz = 0; L = 3
              [[0, 2], [3, 2]],  # lx = 2, ly = 0, lz = 1; L = 3
              [[0], [6]],  # lx = 3, ly = 0, lz = 0; L = 3
              [[1], [7]],  # lx = 0, ly = 3, lz = 0; L = 3
              [[2], [8]]  # lx = 0, ly = 0, lz = 3; L = 3
              ]

    uN = len(unique_js)

    l = 1
    RIJ = np.zeros([len(rij), 9])
    dRIJ = np.zeros([len(rij), 9])

    if L == 1:
        l = 4
        RIJ[:, :3] = rij
        if derivative:
            dRIJ[:, :3] += 1

    elif L == 2:
        l = 10
        RIJ[:, :3] = rij
        RIJ[:, 3:6] = rij * rij
        if derivative:
            dRIJ[:, :3] += 1
            dRIJ[:, 3:6] = 2 * rij

    elif L == 3:
        l = 20
        RIJ[:, :3] = rij
        RIJ[:, 3:6] = rij * rij
        RIJ[:, 6:9] = (rij * rij) * rij

        if derivative:
            dRIJ[:, :3] += 1
            dRIJ[:, 3:6] = 2 * rij
            dRIJ[:, 6:9] = 3 * RIJ[:, 3:6]

    xyz = np.ones([len(rij), 3, l])
    if derivative:
        dxyz = np.zeros([len(rij), uN, 3, l])

    dij_dmlist = dij_dm_list(unique_js, ij_list)  # [j, uN]

    for i in range(1, l):
        xyz[:, L_list[i - 1][0], i] = RIJ[:, L_list[i - 1][1]]
        if derivative:
            dxyz[:, :, L_list[i - 1][0], i] = np.einsum('ij,ik->ijk', dij_dmlist,
                                                        dRIJ[:, L_list[i - 1][1]])  # [j, uN], [j, l] -> [j, uN, l]

    result = xyz[:, 0, :] * xyz[:, 1, :] * xyz[:, 2, :] * normalize[:l]  # [j, l]

    if derivative:
        d_result = np.zeros_like(dxyz)  # [j, uN, 3, l]
        d_result[:, :, 0, :] = np.einsum('ijk,ik->ijk', dxyz[:, :, 0, :], xyz[:, 1, :] * xyz[:, 2, :])
        d_result[:, :, 1, :] = np.einsum('ijk,ik->ijk', dxyz[:, :, 1, :], xyz[:, 0, :] * xyz[:, 2, :])
        d_result[:, :, 2, :] = np.einsum('ijk,ik->ijk', dxyz[:, :, 2, :], xyz[:, 0, :] * xyz[:, 1, :])

        d_result = np.einsum('ijkl,l->ijkl', d_result, normalize[:l])

        return result, d_result
    else:
        return result, None


def dij_dm_list(unique_js, ij_list):
    """Get the sign of the derivative of x-y-z ** lx-ly-lz. 
    
    Parameters
    ----------
    uN: int
        the unique index of the atom that force is acting on.
    ij_list: list
        The list of center atom i w.r.t. the neighbors atom j.

    Returns
    -------
    result: array [j, uN]
        The signs (+ or -) for dXij_dm (YZ) * dYij_dm (XZ) * dZij_dm (XY)
    """
    uN = len(unique_js)
    result = np.zeros([len(ij_list), uN])

    ijm_list = np.zeros([len(ij_list), 3, uN], dtype=int)

    ijm_list[:, -1, :] = unique_js  # np.arange(uN)
    ijm_list[:, :2, :] = np.broadcast_to(ij_list[..., None], ij_list.shape + (uN,))

    arr = (ijm_list[:, 2, :] == ijm_list[:, 0, :])
    result[arr] = -1

    arr = (ijm_list[:, 2, :] == ijm_list[:, 1, :])
    result[arr] = 1

    arr = (ijm_list[:, 0, :] == ijm_list[:, 1, :])  # This condition doesn't seem
    result[arr] = 0  # to contribute anything

    return result  # [j, uN]


def dRij_dRm_norm(Rij, ijm_list):
    """Calculate the derivative of Rij norm w. r. t. atom m. This term affects 
    only on i and j.
    
    Parameters
    ----------
    Rij : array [j, 3]
        The vector distances of atom i to atom j.
    ijm_list: array [j, 3] or [j*k, 3]
        Id list of center atom i, neighbors atom j, and atom m.
    
    Returns
    -------
    dRij_m: array [j, 3]
        The derivative of pair atoms w.r.t. atom m in x, y, z directions.
    """
    dRij_m = np.zeros([len(Rij), 3])

    R1ij = np.linalg.norm(Rij, axis=1).reshape([len(Rij), 1])
    l1 = (ijm_list[:, 2] == ijm_list[:, 0])
    dRij_m[l1, :] = -Rij[l1] / R1ij[l1]
    l2 = (ijm_list[:, 2] == ijm_list[:, 1])
    dRij_m[l2, :] = Rij[l2] / R1ij[l2]
    l3 = (ijm_list[:, 0] == ijm_list[:, 1])
    dRij_m[l3, :] = 0

    return dRij_m


############################# Cutoff Functionals ##############################


def Cosine(dij, rc):
    ids = (dij > rc)
    result = 0.5 * (np.cos(np.pi * dij / rc) + 1.)
    result[ids] = 0.
    return result


def CosinePrime(Rij, Rc):
    # Rij is the norm
    ids = (Rij > Rc)
    result = -0.5 * np.pi / Rc * np.sin(np.pi * Rij / Rc)
    result[ids] = 0

    return result


if __name__ == '__main__':
    from ase.build import bulk

    np.set_printoptions(formatter={'float': '{: 0.4f}'.format})

    Rc = 10
    parameters1 = {'L': 2, 'eta': [0.036, 0.071], 'Rs': [0]}

    # Test for stress
    for a in [5.0]:  # , 5.4, 5.8]:
        si = bulk('Si', 'diamond', a=a, cubic=True)
        cell = si.get_cell()
        cell[0, 1] += 0.1
        si.set_cell(cell)

        bp = EAMD(parameters1, Rc=Rc, derivative=True, stress=True)
        des = bp.calculate(si)

        print("G:", des['x'][0])
        print("GPrime")
        print(des['dxdr'][0:5, :, 2])
        # print(des['rdxdr'][0:8, -1, :, :])
        # pprint(np.einsum('ijklm->klm', des['rdxdr']))



# -*- coding: utf-8 -*-

# @Time    : 2019/10/18 14:27
# @Email   : 986798607@qq.com
# @Software: PyCharm
# @License: BSD 3-Clause

import copy
import os
import re
from itertools import chain
from typing import List

import pandas as pd


class Ele:
    """Element type."""

    def __init__(self, name):
        self.name = [name, ]
        self.num = [1, ]

    def __mul__(self, other):
        if isinstance(other, (float, int)):
            self_new = copy.deepcopy(self)
            self_new.num = [other * _ for _ in self.num]
            return self_new
        else:
            raise TypeError("can not mul {} and  {}".format(self.__str__, other.__str__))

    def __str__(self):
        return "".join(["{}{}".format(i, j) for i, j in zip(self.name, self.num)])

    __repr__ = __str__

    def __rmul__(self, other):
        if isinstance(other, (float, int)):
            self_new = copy.deepcopy(self)
            self_new.num = [other * _ for _ in self.num]
            return self_new
        else:
            raise TypeError("can not mul {} and  {}".format(self.__str__, other.__str__))

    def __add__(self, other):
        if isinstance(other, (float, int)):
            raise TypeError("can not add {} and number {}".format(self.__str__, other))
            pass
        elif isinstance(other, Ele):
            self_new = copy.deepcopy(self)
            for i, j in zip(other.name, other.num):
                if i in self.name:
                    index = self.name.index(i)
                    self_new.num[index] += j
                else:
                    self_new.name.append(i)
                    self_new.num.append(j)
            return self_new

    def __pos__(self):
        return copy.deepcopy(self)

    @property
    def to_dict(self):
        return {i: j for i, j in zip(self.name, self.num)}

    @property
    def to_item(self):
        # print(list(zip(spath.x_name, spath.num)))
        return list(chain.from_iterable(zip(self.name, self.num)))


name_ele = ['H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne', 'Na', 'Mg', 'Al', 'Si', 'P', 'S', 'Cl', 'Ar',
            'K',
            'Ca',
            'Sc', 'Ti', 'V', 'Cr', 'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn', 'Ga', 'Ge', 'As', 'Se', 'Br', 'Kr', 'Rb', 'Sr',
            'Y',
            'Zr', 'Nb', 'Mo', 'Tc', 'Ru', 'Rh', 'Pd', 'Ag', 'Cd', 'In', 'Sn', 'Sb', 'Te', 'I', 'Xe', 'Cs', 'Ba', 'La',
            'Ce', 'Pr', 'Nd', 'Pm', 'Sm', 'Eu', 'Gd', 'Tb', 'Dy', 'Ho', 'Er', 'Tm', 'Yb', 'Lu', 'Hf', 'Ta', 'W', 'Re',
            'Os', 'Ir', 'Pt', 'Au', 'Hg', 'Tl', 'Pb', 'Bi', 'Po', 'At', 'Rn', 'Fr', 'Ra', 'Ac', 'Th', 'Pa', 'U', 'Np',
            'Pu']

for ele_i in name_ele:
    locals()[ele_i] = Ele(ele_i)


def _init_dict(element_n=None):
    if not element_n:
        element_n = len(name_ele)
    dict0 = dict()
    name_list_n = ["Abandon"] + name_ele
    for i in name_list_n[:element_n]:
        dict0[i] = 0
    return dict0


def _fea_dict(ele, element_n=None):
    dict0 = _init_dict(element_n=element_n)
    dicti = copy.copy(dict0)
    num = ele.num
    sname = ele.name
    for j, k in zip(sname, num):
        if j in dicti:
            dicti[j] = k
        else:
            dicti["Abandon"] = 1
    return dicti


def _transform(names, folds_name='folds.csv', expands_name=r'expands.csv'):
    c = []
    for j, i in enumerate(names):
        try:
            com = eval(i)
        except (NameError, SyntaxError) as e:
            print("{}, at {}, which is  {} ".format(e, j, i))
            com = Ele(i)
        c.append(com)
    folds = pd.DataFrame([_i.to_item for _i in c], [str(_) for _ in c])
    expands = pd.DataFrame([_fea_dict(_i) for _i in c], [str(_) for _ in c])
    folds.to_csv(folds_name)
    expands.to_csv(expands_name)


def _pre_add(namei):
    a_ = ['He', 'Li', 'Be', 'Ne', 'Na', 'Mg', 'Al', 'Si', 'Cl', 'Ar', 'Ca', 'Sc', 'Ti', 'Cr', 'Mn', 'Fe', 'Co', 'Ni',
          'Cu', 'Zn', 'Ga', 'Ge', 'As', 'Se', 'Br', 'Kr', 'Rb', 'Sr', 'Zr', 'Nb', 'Mo', 'Tc', 'Ru', 'Rh', 'Pd', 'Ag',
          'Cd', 'In', 'Sn', 'Sb', 'Te', 'Xe', 'Cs', 'Ba', 'La', 'Ce', 'Pr', 'Nd', 'Pm', 'Sm', 'Eu', 'Gd', 'Tb', 'Dy',
          'Ho', 'Er', 'Tm', 'Yb', 'Lu', 'Hf', 'Ta', 'Re', 'Os', 'Ir', 'Pt', 'Au', 'Hg', 'Tl', 'Pb', 'Bi', 'Po', 'At',
          'Rn', 'Fr', 'Ra', 'Ac', 'Th', 'Pa', 'Np', 'Pu']
    b = ['H', 'B', 'C', 'numeric', 'O', 'F', 'K', 'P', 'S', 'V', 'Y', 'I', 'W', 'U']

    num = re.findall(r"\d+\.?\d*", namei)

    namei_new = ""
    namei3 = namei
    j = 0
    while j < len(num):
        s1, s2, namei3 = namei3.partition(num[j])
        namei_new = r"".join([namei_new, s1, "_", s2])
        j += 1
    if namei3:
        namei_new += namei3

    namej = namei_new

    for x in a_:
        namej = re.sub(x, x + '1', namej)
    for x in b:
        namej = re.sub(x, x + '1', namej)
    needsub = re.findall(r'[A-Z]1[][a-z]', namej)
    for i in needsub:
        namej = re.sub(i, i[0] + i[2], namej)
    namej = re.sub(r'1?_', "", namej)

    fin = re.findall(r'\)[A-Z][a-z]*', namej)
    for i in fin:
        namej = re.sub(re.escape(i), ")1" + i[1:], namej, 1)

    return namej


def _bracket_follow(s):
    """
    for situation that the number of () before the ()
    if add this function, make sure the first number below ")" have "+" !!!, if not please add by human.
    """
    pre = re.findall(r'\d+\.?\d*\(.+\)\D*', s)
    long = len(pre)
    while long > 0:
        i = pre[0]
        a_, b, c = i.partition("(")
        d, e, f = c.partition(")")
        res = r"(" + d + r")" + a_ + f
        # s = re.sub(i, res, s, 1)
        s = s.replace(i, res, 1)
        pre = re.findall(r'\d+\.?\d*\(.\)\D*', s)
        long = len(pre)
    return s


def _add_mul(s):
    ret2 = re.findall(r'\d+\.?\d*', s)
    s_new = ""
    s3 = s
    j = 0
    while j < len(ret2):
        s1, s2, s3 = s3.partition(ret2[j])
        s_new = r"".join([s_new, "+", s1, "*", s2])
        j += 1
    if s3:
        s_new += "+" + s3
    s_new = s_new.replace(r"+ +", r"+", 1)
    s_new = s_new.replace(r"+)", r")")
    return s_new


def _substitued(s):
    # print(_bracket_follow.__doc__)
    s = _bracket_follow(s)
    s = _pre_add(s)
    s = _add_mul(s)
    return s


class NameSplit:
    """
    Split the name to table, return expand.csv and folds.csv.

    Notes
    -----
        Make sure the number below the element or (),
        For situation that the number of () before the (), such as "0.9(BaZr)+0.1(BaZrO3)".

        Please set the bracket_follow=True,
        and for this situation make sure the first number below ``)`` have ``+`` !!!.

    Examples
    --------
    >>> from featurebox.data.namespilt import NameSplit
    >>> import os
    >>> os.chdir(r'.')
    >>> name = ['(Ti1.24La3)2',"((Ti1.24)2P2)1H0.2", "((Ti1.24)2)1H0.2",
    ... "((Ti1.24))1H0.2", "((Ti)2P2)1H0.2",  "((Ti))1H0.2"]
    >>> NSp = NameSplit()
    >>> NSp.transform(name)
    ...
    """

    def __init__(self, bracket_follow: bool = False):
        self.brack = bracket_follow

    def _substitute(self, s: List):
        if self.brack:
            print(_bracket_follow.__doc__)
            s = [_bracket_follow(si) for si in s]
        s = [_pre_add(si) for si in s]
        s = [_add_mul(si) for si in s]
        return s

    def transform(self, names: List[str], folds_name: str = 'folds.csv', expands_name: str = r'expands.csv'):
        """

        Parameters
        ----------
        names:list of str
            composition names list.
        folds_name:str
            return name, default is 'folds.csv'
        expands_name:str
            return name, default is 'expands.csv'

        Returns
        -------
        None
            return tables.csv.
        """
        names = self._substitute(names)
        _transform(names, folds_name, expands_name)


if __name__ == '__main__':
    a = [
        '(BaZr0.2Ti0.8O3)0.9(Ba0.7Ca0.3TiO3)0.1',
        '(Ba0.95Ca0.05)+(Zr0.1Ti0.9)O3',
        '(Ba(Zr0.2Ti0.8)O3)0.5 + ((Ba0.7Ca0.3)TiO3)0.5',
        'Ba0.95Sr0.05+(Ti0.95Zr0.05)0.975Sn0.025O3',
        "Ba0.95Sr0.05(Ti0.95Zr0.05)O3",
        "Ba0.9Sr0.1(Ti0.9Zr0.1)0.95Sn0.05O3",
        '(Ti1.24La3)2',
        "((Ti1.24)2P2)1H0.2", "((Ti1.24)2)1H0.2", "((Ti1.24))1H0.2",
        r"(TiLa)H2", r"(TiLa)0.2H2",
        r"(TiLa)0.2", '(TiLa)',
        '(TiLa3)', '(TiLa)', '(TiLa3)2',
        "((Ti)2P2)1H0.2", "((Ti)2)1H0.2", "((Ti))1H0.2"
    ]
    os.chdir(r'.')

    NS = NameSplit()

    NS.transform(a)



import unittest

import pandas as pd

from featurebox.data.check_data import CheckElements
from featurebox.featurizers.base_graph import CrystalGraph


class TestGraph2(unittest.TestCase):
    def setUp(self) -> None:

        self.data = pd.read_pickle("data_structure.pkl_pd")
        self.data0 = self.data[0]
        self.data0_3 = self.data[:3]
        self.data0_3 = self.data[:3]
        ce = CheckElements.from_pymatgen_structures()

        self.data0_checked = ce.check(self.data)[:10]

    def test_CrystalGraph(self):
        for i in self.data0_checked:
            sg1 = CrystalGraph(nn_strategy="SOAP", bond_generator="BaseDesGet", cutoff=None)
            s12 = sg1(i)
            # print(s12)
            print(s12["bond"].shape[-2])
            print(s12["bond"].shape[-1])

    def test_CrystalGraph2(self):
        for i in self.data0_checked:
            sg1 = CrystalGraph(nn_strategy="EAMD", bond_generator="BaseDesGet", cutoff=None)
            s12 = sg1(i)
            # print(s12)
            print(s12["bond"].shape[-2])
            print(s12["bond"].shape[-1])


if __name__ == '__main__':
    unittest.main()



# Running Bispectrum Calculations with LAMMPS
# Pymatgen based

import os
import subprocess

import numpy as np
from pymatgen.io.lammps.data import LammpsData


def make_js(twojmax, diagonal):
    js = []
    for j1 in range(0, twojmax + 1):
        if diagonal == 2:
            js.append([j1, j1, j1])
        elif diagonal == 1:
            for j in range(0, min(twojmax, 2 * j1) + 1, 2):
                js.append([j1, j1, j])
        elif diagonal == 0:
            for j2 in range(0, j1 + 1):
                for j in range(j1 - j2, min(twojmax, j1 + j2) + 1, 2):
                    js.append([j1, j2, j])
        elif diagonal == 3:
            for j2 in range(0, j1 + 1):
                for j in range(j1 - j2, min(twojmax, j1 + j2) + 1, 2):
                    if j >= j1:
                        js.append([j1, j2, j])
    return js


# Perhaps, this class can be optimized further more if we don't include stress
# calculations if there is no needed.
class Bispectrum:
    """This class prepares a lammps input file and calls the lammps executable
    to calculate bispectrum coefficients of a given structure.
    
    Parameters
    ----------
    structure: object
        Pymatgen crystal structure object.
    rcutfac: float
        Scale factor applied to all cutoff radii.
    element_profile: dict
        Elemental descriptions of each atom type in the structure.
        i.e. dict(Na=dict(r=0.3, w=0.9), Cl=dict(r=0.7, w=3.0)).
    twojmax: int
        Band limit for bispectrum components.
    diagonal: int
        diagonal value = 0 or 1 or 2 or 3.
        0 = all j1, j2, j <= twojmax, j2 <= j1
        1 = subset satisfying j1 == j2
        2 = subset satisfying j1 == j2 == j3
        3 = subset satisfying j2 <= j1 <= j
    rfac0: float
        Parameter in distance to angle conversion (0 < rcutfac < 1).
        Default value: 0.99363.
    rmin0: float
        Parameter in distance to angle conversion.
        Default value: 0.
    """

    def __init__(self, structure, rcutfac, element_profile, twojmax,
                 diagonal=3, rfac0=0.99363, rmin0=0.):
        self.exe = 'lmp_serial'  # define your LAMMPS executable here
        self.pre_cmds = ['units metal',
                         'atom_style charge',
                         'box tilt large',
                         'read_data data.0',
                         'pair_style lj/cut 10',
                         'pair_coeff * * 1 1']
        self.compute_cmds = ['compute sna all sna/atom ',
                             'compute snad all snad/atom ',
                             'compute snav all snav/atom ',
                             'dump 1 all custom 1 dump.element element',
                             'dump 2 all custom 1 dump.sna c_sna[*]',
                             'dump 3 all custom 1 dump.snad c_snad[*]',
                             'dump 4 all custom 1 dump.snav c_snav[*]',
                             'dump_modify 1 element ']
        self.post_cmds = ['run 0']
        self.input_file = 'in.sna'

        self.structure = structure
        self.rcutfac = rcutfac
        self.twojmax = twojmax
        self.elements = element_profile.keys()
        self.ele = []
        self.Rs = []
        self.Ws = []
        for key, value in element_profile.items():
            self.ele.append(key)
            self.Rs.append(element_profile[key]['r'])
            self.Ws.append(element_profile[key]['w'])

        self.rfac0 = rfac0
        self.rmin0 = rmin0
        assert diagonal in range(4), \
            'Invalid diagonal style, must be 0, 1, 2, or 3.'
        self.diagonal = diagonal

        self.calculate()

        os.remove("data.0")
        os.remove("in.sna")
        os.remove("log.lammps")

    def calculate(self):
        """Call the LAMMPS executable to compute bispectrum coefficients."""
        data = self.get_dump(self.structure, self.elements)
        data.write_file('data.0')
        self.get_lammps_input(self.input_file)
        p = subprocess.Popen([self.exe, '-in', self.input_file],
                             stdout=subprocess.PIPE)
        stdout = p.communicate()[0]
        rc = p.returncode
        if rc != 0:
            error_msg = 'LAMMPS exited with return code %d' % rc
            msg = stdout.decode("utf-8").split('\n')[:-1]
            try:
                error = [i for i, m in enumerate(msg)
                         if m.startswith('ERROR')][0]
                error_msg += ', '.join([e for e in msg[error:]])
            except:
                error_msg += msg[-1]
            print(error_msg)
            # raise RuntimeError(error_msg)

    def get_dump(self, structure, elements):
        """Convert Pymatgen structure object to LAMMPS dump file."""
        data = LammpsData.from_structure(structure, elements)

        return data

    def get_lammps_input(self, input_file):
        """Create LAMMPS input file."""
        sna = f"1 {self.rfac0} {self.twojmax} "
        for R in self.Rs:
            R *= self.rcutfac
            sna += f"{R} "
        for W in self.Ws:
            sna += f"{W} "
        sna += f"diagonal {self.diagonal} rmin0 {self.rmin0} "

        self.compute_cmds[0] += sna + "bzeroflag 0"
        self.compute_cmds[1] += sna + "quadraticflag 0"
        self.compute_cmds[2] += sna + "quadraticflag 0"

        for el in self.ele:
            self.compute_cmds[-1] += f"{el} "

        self.CMDS = self.pre_cmds + self.compute_cmds + self.post_cmds

        with open(input_file, 'w') as f:
            for line in self.CMDS:
                f.write("%s\n" % line)


class Assembler:
    """
    A class to gather all bispectrum components, including force and stress
    bispectrum. 
    
    Parameters
    ----------
    atom_type: list of str
        String of all atom types in the structure, i.e. ['Na', 'Cl']
    volume: float
        Volume of a crystal structure. Volume is used to convert the stress
        bispectrum components from eV to GPa.
    force: bool
        If true, return force bispectrum components.
    stress: bool
        If true, return stress bispectrum components
    """

    def __init__(self, atom_type, volume, force=True, stress=False):
        self.d_bispectrum_coefficient = {}
        elements = self._read_dump("dump.element", dtype='str')

        # sna
        bias_weight = []
        for _ in range(len(atom_type)):
            bias_weight.append([1.0 / len(atom_type)])

        sna = self._read_dump("dump.sna")
        self.sna = []
        for atom in atom_type:
            sum = np.zeros(len(sna[0]))
            for i, ele in enumerate(elements):
                if ele == atom:
                    sum += sna[i]
            self.sna.append(sum / (i + 1))
        self.sna = np.hstack((bias_weight, self.sna))
        self.sna = [np.ravel(self.sna)]

        all = self.sna
        # print(f"This is self.sna:\n{self.sna}")

        # snad
        if force == True:
            snad = self._read_dump("dump.snad")
            snad = np.split(np.asarray(snad), len(atom_type), axis=1)
            depth, rows, columns = np.shape(snad)
            x, y, z = int(columns / 3), int(columns * 2 / 3), columns

            self.snad = []
            for i in range(len(atom_type)):
                temp = []
                for j in range(len(elements)):
                    temp.append(snad[i][j][0:x])  # x-direction
                    temp.append(snad[i][j][x:y])  # y-direction
                    temp.append(snad[i][j][y:z])  # z-direction
                if self.snad == []:
                    self.snad = np.hstack((np.zeros((len(temp), 1)), temp))
                else:
                    temp = np.hstack((np.zeros((len(temp), 1)), temp))
                    self.snad = np.hstack((self.snad, temp))

            all = np.concatenate((self.sna, self.snad))
            # print(f"This is self.snad:\n{self.snad}")

        # snav
        if stress == True:
            snav = self._read_dump("dump.snav")
            snav = np.asarray(snav)
            snav = np.split(snav.sum(axis=0), len(atom_type))

            self.snav = []
            for i in range(len(atom_type)):
                temp = np.reshape(snav[i], (6, len(sna[0])))
                if self.snav == []:
                    self.snav = np.hstack((np.zeros((len(temp), 1)), temp))
                else:
                    temp = np.hstack((np.zeros((len(temp), 1)), temp))
                    self.snav = np.hstack((self.snav, temp))
            self.snav = self.snav / volume * 160.21766208  # eV to GPa
            all = np.concatenate((all, self.snav))
            # print(f"This is self.snav:\n{self.snav}")

        # print(f"This is all:\n{all}")
        self.bispectrum_coefficients = all

        # Remove the dump files after usage due to write error.
        os.remove("dump.element")
        os.remove("dump.sna")
        os.remove("dump.snad")
        os.remove("dump.snav")

    @staticmethod
    def _read_dump(filename, dtype='float'):
        arr = []

        file = open(filename)
        lines = file.readlines()
        n_lines = int(lines[3])
        goods = lines[9:]

        for i in range(n_lines):
            l = goods[i].split()

            if dtype == 'float':
                l = np.asarray([float(z) for z in l])
                arr.append(l)

            else:
                l = l[0]
                arr.append(l)

        return arr
